# -*- coding: utf-8 -*-

# Copyright © 2012-2015 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function, unicode_literals

import unicodedata
import time

import logging
_logger = logging.getLogger(__name__)

from Onboard.Version import require_gi_versions
require_gi_versions()
try:
    from gi.repository import Atspi
except ImportError as e:
    pass

from Onboard.AtspiStateTracker import AtspiStateTracker, AtspiStateType
from Onboard.TextDomain        import TextDomains
from Onboard.TextChanges       import TextChanges, TextSpan
from Onboard.utils             import KeyCode
from Onboard.Timer             import Timer
from Onboard                   import KeyCommon

### Config Singleton ###
from Onboard.Config import Config
config = Config()


class TextContext:
    """
    Keep track of the current text context and intecept typed key events.
    """

    def cleanup(self):
        pass

    def reset(self):
        pass

    def can_insert_text(self):
        return NotImplementedError()

    def insert_text(self, offset, text):
        return NotImplementedError()

    def insert_text_at_caret(self, text):
        return NotImplementedError()

    def delete_text(self, offset, length=1):
        return NotImplementedError()

    def delete_text_before_caret(self, length=1):
        return NotImplementedError()

    def get_context(self):
        raise NotImplementedError()

    def get_line(self):
        raise NotImplementedError()

    def get_line_caret_pos(self):
        raise NotImplementedError()

    def get_changes(self):
        raise NotImplementedError()

    def clear_changes(self):
        raise NotImplementedError()


class AtspiTextContext(TextContext):
    """
    Keep track of the current text context with AT-SPI
    """

    _state_tracker = AtspiStateTracker()

    def __init__(self, wp):
        self._wp = wp
        self._accessible = None
        self._can_insert_text = False

        self._text_domains = TextDomains()
        self._text_domain = self._text_domains.get_nop_domain()

        self._changes = TextChanges()
        self._entering_text = False
        self._text_changed = False

        self._context = ""
        self._line = ""
        self._line_caret = 0
        self._selection_span = TextSpan()
        self._begin_of_text = False        # context starts at begin of text?
        self._begin_of_text_offset = None  # offset of text begin

        self._pending_separator_span = None
        self._last_text_change_time = 0
        self._last_caret_move_time = 0
        self._last_caret_move_position = 0

        self._last_context = None
        self._last_line = None

        self._update_context_timer = Timer()
        self._update_context_delay_normal = 0.01
        self._update_context_delay = self._update_context_delay_normal

    def cleanup(self):
        self._register_atspi_listeners(False)

    def enable(self, enable):
        self._register_atspi_listeners(enable)

    def get_text_domain(self):
        return self._text_domain

    def set_pending_separator(self, separator_span=None):
        """ Remember this separator span for later insertion. """
        if self._pending_separator_span is not separator_span:
            self._pending_separator_span = separator_span

    def get_pending_separator(self):
        """ Return current pending separator span or None """
        return self._pending_separator_span

    def get_context(self):
        """
        Returns the predictions context, i.e. some range of
        text before the caret position.
        """
        if self._accessible is None:
            return ""

        # Don't update suggestions in scrolling terminals
        if self._entering_text or \
           not self._text_changed or \
           self.can_suggest_before_typing():
            return self._context

        return ""

    def get_bot_context(self):
        """
        Returns the predictions context with
        begin of text marker (at text begin).
        """
        context = ""
        if self._accessible:
            context = self.get_context()

            # prepend domain specific begin-of-text marker
            if self._begin_of_text:
                marker = self.get_text_begin_marker()
                if marker:
                    context = marker + " " + context

        return context

    def get_pending_bot_context(self):
        """
        Context including bot marker and pending separator.
        """
        context = self.get_bot_context()
        if self._pending_separator_span is not None:
            context += self._pending_separator_span.get_span_text()
        return context

    def get_line(self):
        return self._line \
               if self._accessible else ""

    def get_line_caret_pos(self):
        return self._line_caret \
               if self._accessible else 0

    def get_line_past_caret(self):
        return self._line[self._line_caret:] \
               if self._accessible else ""

    def get_selection_span(self):
        return self._selection_span \
               if self._accessible else None

    def get_span_at_caret(self):
        if not self._accessible:
            return None

        span = self._selection_span.copy()
        span.length = 0
        return span

    def get_caret(self):
        return self._selection_span.begin() \
            if self._accessible else 0

    def get_character_extents(self, offset):
        accessible = self._accessible
        if accessible:
            return self._state_tracker.get_accessible_character_extents(
                accessible, offset)
        else:
            return None

    def get_text_begin_marker(self):
        domain = self.get_text_domain()
        if domain:
            return domain.get_text_begin_marker()
        return ""

    def can_record_insertion(self, accessible, pos, length):
        domain = self.get_text_domain()
        if domain:
            return domain.can_record_insertion(accessible, pos, length)
        return True

    def can_suggest_before_typing(self):
        domain = self.get_text_domain()
        if domain:
            return domain.can_suggest_before_typing()
        return True

    def can_auto_punctuate(self):
        domain = self.get_text_domain()
        if domain:
            return domain.can_auto_punctuate(self._begin_of_text)
        return False

    def get_begin_of_text_offset(self):
        return self._begin_of_text_offset \
               if self._accessible else None

    def get_changes(self):
        return self._changes

    def has_changes(self):
        """ Are there any changes to learn? """
        return not self._changes.is_empty()

    def clear_changes(self):
        self._changes.clear()

    def can_insert_text(self):
        """
        Can delete or insert text into the accessible?
        """
        #return False # support for inserting is spotty: not in firefox, terminal
        return bool(self._accessible) and self._can_insert_text

    def delete_text(self, offset, length=1):
        """ Delete directly, without going through faking key presses. """
        self._accessible.delete_text(offset, offset + length)

    def delete_text_before_caret(self, length=1):
        """ Delete directly, without going through faking key presses. """
        offset = self._accessible.get_caret_offset()
        self.delete_text(offset - length, length)

    def insert_text(self, offset, text):
        """
        Insert directly, without going through faking key presses.
        """
        self._accessible.insert_text(offset, text, -1)

        # Move the caret after insertion if the accessible itself
        # hasn't done so already. This assumes the insertion begins at
        # the current caret position, which always happens to be the case
        # currently.
        # Only the nautilus rename text entry appears to need this.
        offset_before = offset
        offset_after = self._accessible.get_caret_offset()
        if text and offset_before == offset_after:
            self._accessible.set_caret_offset(offset_before + len(text))

    def insert_text_at_caret(self, text):
        """
        Insert directly, without going through faking key presses.
        Fails for terminal and firefox, unfortunately.
        """
        offset = self._accessible.get_caret_offset()
        self.insert_text(offset, text)

    def _register_atspi_listeners(self, register=True):
        st = self._state_tracker
        if register:
            st.connect("text-entry-activated", self._on_text_entry_activated)
            st.connect("text-changed", self._on_text_changed)
            st.connect("text-caret-moved", self._on_text_caret_moved)
            # st.connect("key-pressed", self._on_atspi_key_pressed)
        else:
            st.disconnect("text-entry-activated", self._on_text_entry_activated)
            st.disconnect("text-changed", self._on_text_changed)
            st.disconnect("text-caret-moved", self._on_text_caret_moved)
            # st.disconnect("key-pressed", self._on_atspi_key_pressed)

    def get_accessible_capabilities(self, accessible, **kwargs):
        can_insert_text = False
        attributes = kwargs.get("attributes", {})
        interfaces = kwargs.get("interfaces", [])

        if accessible:

            # Can insert text via Atspi?
            # Advantages:
            # - faster, no individual key presses
            # - trouble-free insertion of all unicode characters
            if "EditableText" in interfaces:
                # Support for atspi text insertion is spotty.
                # Firefox, LibreOffice Writer, gnome-terminal don't support it,
                # even if they claim to implement the EditableText interface.

                # Allow direct text insertion for gtk widgets
                if self._state_tracker.is_toolkit_gtk3(attributes):
                    can_insert_text = True

        return can_insert_text

    def _on_text_entry_activated(self, accessible):
        # old text_domain still valid here
        self._wp.on_text_entry_deactivated()

        # keep track of the active accessible asynchronously
        self._accessible = accessible
        self._entering_text = False
        self._text_changed = False

        # select text domain matching this accessible
        state = self._state_tracker.get_state() \
                if self._accessible else {}
        self._text_domain = self._text_domains.find_match(**state)
        self._text_domain.init_domain()

        # determine capabilities of this accessible
        self._can_insert_text = \
            self.get_accessible_capabilities(accessible, **state)

        # log accessible info
        if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
            log = _logger.atspi
            log("-" * 70)
            log("Accessible focused: ")
            indent = " " * 4
            if self._accessible:
                state = self._state_tracker.get_state()
                for key, value in sorted(state.items()):
                    msg = str(key) + "="
                    if key == "state-set":
                        msg += repr(AtspiStateType.to_strings(value))
                    elif hasattr(value, "value_name"):  # e.g. role
                        msg += value.value_name
                    else:
                        msg += repr(value)
                    log(indent + msg)
                log(indent + "text_domain: {}"
                    .format(self._text_domain and
                            type(self._text_domain).__name__))
                log(indent + "can_insert_text: {}"
                    .format(self._can_insert_text))
            else:
                log(indent + "None")

        self._update_context()

        self._wp.on_text_entry_activated()

    def _on_text_changed(self, event):
        insertion_span = self._record_text_change(event.pos,
                                                  event.length,
                                                  event.insert)
        # synchronously notify of text insertion
        if insertion_span:
            try:
                caret_offset = self._accessible.get_caret_offset()
            except:  # gi._glib.GError
                pass
            else:
                self._wp.on_text_inserted(insertion_span, caret_offset)

        self._last_text_change_time = time.time()
        self._update_context()

    def _on_text_caret_moved(self, event):
        self._last_caret_move_time = time.time()
        self._last_caret_move_position = event.caret
        self._update_context()
        self._wp.on_text_caret_moved()

    def _on_atspi_key_pressed(self, event):
        """ disabled, Francesco didn't receive any AT-SPI key-strokes. """
        # keycode = event.hw_code # uh oh, only keycodes...
        #                         # hopefully "c" doesn't move around a lot.
        # modifiers = event.modifiers
        # self._handle_key_press(keycode, modifiers)

    def on_onboard_typing(self, key, mod_mask):
        if key.is_text_changing():
            keycode = 0
            if key.is_return():
                keycode = KeyCode.KP_Enter
            else:
                label = key.get_label()
                if label == "C" or label == "c":
                    keycode = KeyCode.C

            self._handle_key_press(keycode, mod_mask)

    def _handle_key_press(self, keycode, modifiers):
        if self._accessible:
            domain = self.get_text_domain()
            if domain:
                self._entering_text, end_of_editing = \
                    domain.handle_key_press(keycode, modifiers)

                if end_of_editing is True:
                    self._wp.commit_changes()
                elif end_of_editing is False:
                    self._wp.discard_changes()

    def _record_text_change(self, pos, length, insert):
        accessible = self._accessible

        insertion_span = None
        char_count = None
        if accessible:
            try:
                char_count = accessible.get_character_count()
            except:     # gi._glib.GError: The application no longer exists
                        # when closing a tab in gnome-terminal.
                char_count = None

        if char_count is not None:
            # record the change
            spans_to_update = []

            if insert:
                if self._entering_text and \
                   self.can_record_insertion(accessible, pos, length):
                    if self._wp.is_typing() or length < 30:
                        # Remember all of the insertion, might have been
                        # a pressed snippet or wordlist button.
                        include_length = -1
                    else:
                        # Remember only the first few characters.
                        # Large inserts can be paste, reload or scroll
                        # operations. Only learn the first word of these.
                        include_length = 2

                    # simple span for current insertion
                    begin = max(pos - 100, 0)
                    end = min(pos + length + 100, char_count)
                    text = self._state_tracker.get_accessible_text(accessible,
                                                                   begin, end)
                    if text is not None:
                        insertion_span = TextSpan(pos, length, text, begin)
                else:
                    # Remember nothing, just update existing spans.
                    include_length = None

                spans_to_update = self._changes.insert(pos, length,
                                                       include_length)

            else:
                spans_to_update = self._changes.delete(pos, length,
                                                       self._entering_text)

            # update text of all modified spans
            for span in spans_to_update:
                # Get some more text around the span to hopefully
                # include whole words at beginning and end.
                begin = max(span.begin() - 100, 0)
                end = min(span.end() + 100, char_count)
                span.text = Atspi.Text.get_text(accessible, begin, end)
                span.text_pos = begin

        self._text_changed = True

        return insertion_span

    def set_update_context_delay(self, delay):
        self._update_context_delay = delay

    def reset_update_context_delay(self):
        self._update_context_delay = self._update_context_delay_normal

    def _update_context(self):
        self._update_context_timer.start(self._update_context_delay,
                                         self.on_text_context_changed)

    def on_text_context_changed(self):
        # Clear pending separator when the user clicked to move
        # the cursor away from the separator position.
        if self._pending_separator_span:
            # Lone caret movement, no recent text change?
            if self._last_caret_move_time - self._last_text_change_time > 1.0:
                # Away from the separator?
                if self._last_caret_move_position != \
                   self._pending_separator_span.begin():
                    self.set_pending_separator(None)

        result = self._text_domain.read_context(self._accessible)
        if result is not None:
            (self._context,
             self._line,
             self._line_caret,
             self._selection_span,
             self._begin_of_text,
             self._begin_of_text_offset) = result

            # make sure to include bot-markers and pending separator
            context = self.get_pending_bot_context()
            change_detected = (self._last_context != context or
                               self._last_line != self._line)
            if change_detected:
                self._last_context = context
                self._last_line    = self._line

            self._wp.on_text_context_changed(change_detected)

        return False


class InputLine(TextContext):
    """
    Track key presses ourselves.
    Advantage: Doesn't require AT-SPI
    Problems:  Misses key repeats,
               Doesn't know about keymap translations before events are
               delivered to their destination, i.e records wrong key
               strokes when changing keymaps.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.line = ""
        self.caret = 0
        self.valid = True

        self.word_infos = {}

    def is_valid(self):
        return self.valid

    def is_empty(self):
        return len(self.line) == 0

    def insert(self, s):
        self.line   = self.line[:self.caret] + s + self.line[self.caret:]
        self.move_caret(len(s))

    def delete_left(self, n=1):     # backspace
        self.line = self.line[:self.caret - n] + self.line[self.caret:]
        self.move_caret(-n)

    def delete_right(self, n=1):    # delete
        self.line = self.line[:self.caret] + self.line[self.caret + n:]

    def move_caret(self, n):
        self.caret += n

        # moving into unknown territory -> suggest reset
        if self.caret < 0:
            self.caret = 0
            self.valid = False
        if self.caret > len(self.line):
            self.caret = len(self.line)
            self.valid = False

    def get_context(self):
        return self.line[:self.caret]

    def get_line(self):
        return self.line

    def get_line_caret_pos(self):
        return self.caret

    @staticmethod
    def is_printable(char):
        """
        True for printable keys including whitespace as defined for isprint().
        """
        if char == "\t":
            return True
        return not unicodedata.category(char) in ('Cc', 'Cf', 'Cs', 'Co',
                                                  'Cn', 'Zl', 'Zp')

    def track_sent_key(self, key, mods):
        """
        Sync input_line with single key presses.
        WORD_ACTION and MACRO_ACTION do this in press_key_string.
        """
        end_editing = False

        if config.wp.stealth_mode:
            return True

        id = key.id.upper()
        char = key.get_label()
        if char is None or len(char) > 1:
            char = ""

        if key.action_type == KeyCommon.WORD_ACTION:
            pass  # don't reset input on word insertion

        elif key.action_type == KeyCommon.MODIFIER_ACTION:
            pass  # simply pressing a modifier shouldn't stop the word

        elif key.action_type == KeyCommon.BUTTON_ACTION:
            pass

        elif key.action_type == KeyCommon.KEYSYM_ACTION:
            if id == 'ESC':
                self.reset()
            end_editing = True

        elif key.action_type == KeyCommon.KEYPRESS_NAME_ACTION:
            if id == 'DELE':
                self.delete_right()
            elif id == 'LEFT':
                self.move_caret(-1)
            elif id == 'RGHT':
                self.move_caret(1)
            else:
                end_editing = True

        elif key.action_type == KeyCommon.KEYCODE_ACTION:
            if id == 'RTRN':
                char = "\n"
            elif id == 'SPCE':
                char = " "
            elif id == 'TAB':
                char = "\t"

            if id == 'BKSP':
                self.delete_left()
            elif self.is_printable(char):
                if mods[4]:  # ctrl+key press?
                    end_editing = True
                else:
                    self.insert(char)
            else:
                end_editing = True
        else:
            end_editing = True

        if not self.is_valid():  # caret moved outside known range?
            end_editing = True

        # print end_editing,"'%s' " % self.line, self.caret
        return end_editing




# -*- coding: utf-8 -*-

# Copyright © 2007-2009 Chris Jones <tortoise@tortuga>
# Copyright © 2008, 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
# Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
# Copyright © 2011-2015 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function, unicode_literals

import sys
import os
import time
import re
import colorsys
import gettext
import subprocess
from math import pi, sin, cos, sqrt, log
from contextlib import contextmanager

import logging
from functools import reduce

from gi.repository import GLib

_logger = logging.getLogger("utils")


# keycodes
class KeyCode:
    Return   = 36
    KP_Enter = 104
    C        = 54


class Modifiers:
    # 1      2     4    8    16     32    64     128
    SHIFT, CAPS, CTRL, ALT, NUMLK, MOD3, SUPER, ALTGR = \
               (1<<bit for bit in range(8))

# modifiers affecting labels
LABEL_MODIFIERS = Modifiers.SHIFT | \
                  Modifiers.CAPS | \
                  Modifiers.NUMLK | \
                  Modifiers.ALTGR

modifiers = {"shift":1,
             "caps":2,
             "control":4,
             "mod1":8,   # Left Alt
             "mod2":16,  # NumLk
             "mod3":32,
             "mod4":64,  # Super
             "mod5":128, # Alt Gr
            }

modGroups = {"SHIFT" : ["LFSH"],
             "CTRL"  : ["LCTL"],
            }

modList = [["LWIN", ("Win",64)],
           ["RTSH", ("⇧", 1)],
           ["LFSH", ("⇧", 1)],
           ["RALT", ("Alt Gr", 128)],
           ["LALT", ("Alt", 8)],
           ["RCTL", ("Ctrl", 4)],
           ["LCTL", ("Ctrl", 4)],
           ["CAPS", ("CAPS", 2)],
           ["NMLK", ("Nm\nLk",16)]]

modDic = dict(modList)

otherDic = {"RWIN" : "Win",
            "MENU" : "Menu",
            "BKSP" : "⇦",
            "RTRN" : "Return",
            "TAB" : "Tab",
            "INS":"Ins",
            "HOME":"Hm",
            "PGUP": "Pg\nUp",
            "DELE":"Del",
            "END":"End",
            "PGDN":"Pg\nDn",
            "UP":  "↑",
            "DOWN":"↓",
            "LEFT" : "←",
            "RGHT" : "→",
            "KP0" : "0",
            "KP1" : "1",
            "KP2" : "2",
            "KP3" : "3",
            "KP4" : "4",
            "KP5" : "5",
            "KP6" : "6",
            "KP7" : "7",
            "KP8" : "8",
            "KP9" : "9",
            "KPDL":"Del",
            "KPEN": "Ent" }

funcKeys = (("ESC",65307),
            ("F1",65470),
            ("F2",65471),
            ("F3",65472),
            ("F4", 65473),
            ("F5", 65474),
            ("F6",65475),
            ("F7",65476),
            ("F8",65477),
            ("F9",65478),
            ("F10",65479),
            ("F11", 65480),
            ("F12", 65481),
            ("Prnt", 65377),
            ("Scroll", 65300),
            ("Pause", 65299))

keysyms = {"space" : 0x0020,
           "kp_space" : 65408,
           "insert" : 0xff63,
           "home" : 0xff50,
           "page_up" : 0xff55,
           "page_down" : 0xff56,
           "end" :0xff57,
           "delete" : 0xffff,
           "return" : 65293,
           "backspace" : 65288,
           "left" : 0xff51,
           "up" : 0xff52,
           "right" : 0xff53,
           "down" : 0xff54,}

def get_keysym_from_name(name):
    return keysyms[name]

def parse_key_combination(combo, avaliable_key_ids = None):
    """
    Parses a key combination into a list of modifier masks and key_ids.
    The key-id part of the combo may contain a regex pattern.

    Doctests:

    # modifiers
    >>> parse_key_combination(["TAB"], ["TAB"])
    [('TAB', 0)]
    >>> parse_key_combination(["LALT", "TAB"], ["TAB"])
    [('TAB', 8)]
    >>> parse_key_combination(["LALT", "LFSH", "TAB"], ["TAB"])
    [('TAB', 9)]
    >>> parse_key_combination(["LWIN", "RTSH", "LFSH", "RALT", "LALT", "RCTL", "LCTL", "CAPS", "NMLK", "TAB"], ["TAB"])
    [('TAB', 223)]

    # modifier groups
    >>> parse_key_combination(["CTRL", "SHIFT", "TAB"], ["TAB"])
    [('TAB', 5)]

    # regex
    >>> parse_key_combination(["F\d+"], ["TAB", "F1", "F2", "F3", "F9"])
    [('F1', 0), ('F2', 0), ('F3', 0), ('F9', 0)]
    """
    modifiers = combo[:-1]
    key_pattern = combo[-1]

    # find modifier mask
    mod_mask = parse_modifier_strings(modifiers)
    if mod_mask is None:
        return None

    # match regex key id with all available ids
    results = []
    pattern = re.compile(key_pattern)
    for key_id in avaliable_key_ids:
        match = pattern.match(key_id)
        if match and match.group() == key_id:
            results.append((key_id, mod_mask))

    return results

def parse_modifier_strings(modifiers):
    """ Build modifier mask from modifier strings. """
    mod_mask = 0
    for modifier in modifiers:
        m = modDic.get(modifier)
        if not m is None:
            mod_mask |= m[1]
        else:
            group = modGroups.get(modifier)
            if not group is None:
                for mod in group:
                    mod_mask |= modDic[mod][1]
            else:
                _logger.warning("unrecognized modifier '{}'; try one of {}" \
                                .format(modifier, ",".join(m[0] for m in modList)))
                mod_mask = None
                break

    return mod_mask

def run_script(script):
    a =__import__(script)
    a.run()

def toprettyxml(domdoc):
    ugly_xml = domdoc.toprettyxml(indent='  ')
    # Join lines with text elements with their tag lines
    pattern = re.compile('>\n\s+([^<>\s].*?)\n\s+</', re.DOTALL)
    pretty_xml = pattern.sub('>\g<1></', ugly_xml)

    # Work around http://bugs.python.org/issue5752
    pretty_xml = re.sub(
           '"[^"]*"',
           lambda m: m.group(0).replace("\n", "&#10;"),
           pretty_xml)

    # remove empty lines
    pretty_xml = os.linesep.join( \
                    [s for s in pretty_xml.splitlines() if s.strip()])
    return pretty_xml


def dec_to_hex_colour(dec):
    hexString = hex(int(255*dec))[2:]
    if len(hexString) == 1:
        hexString = "0" + hexString

    return hexString

def xml_get_text(dom_node, tag_name):
    """ extract text from a dom node """
    nodelist = dom_node.getElementsByTagName(tag_name)
    if not nodelist:
        return None
    rc = []
    for node in nodelist[0].childNodes:
        if node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
    return ''.join(rc).strip()

def matmult(m, v):
    """ Matrix-vector multiplication """
    nrows = len(m)
    w = [None] * nrows
    for row in range(nrows):
        w[row] = reduce(lambda x,y: x+y, list(map(lambda x,y: x*y, m[row], v)))
    return w

def hexstring_to_float(hexString):
    return float(int(hexString, 16))

def hexcolor_to_rgba(color):
    """
    convert '#rrggbb' or '#rrggbbaa' to (r, g, b, a)

    Doctests:
    >>> def test(color):
    ...     rgba = hexcolor_to_rgba(color)
    ...     if rgba is None:
    ...         print(repr(rgba))
    ...     else:
    ...         print(repr([round(c, 2) for c in rgba]))

    >>> test("#1a2b3c")
    [0.1, 0.17, 0.24, 1.0]

    >>> test("#1a2b3c4d")
    [0.1, 0.17, 0.24, 0.3]

    >>> test("")
    None

    >>> test("1a2b3c")
    None

    >>> test("1a2b3c4d")
    None

    >>> test("#1a2b3c4dx")
    None

    >>> test("#1a2b3cx")
    None

    >>> test("#1a2bx")
    None

    >>> test("#1aXb3c4d")
    None
    """
    rgba = None
    n = len(color)
    if n == 7 or n == 9:
        try:
            rgba = [hexstring_to_float(color[1:3])/255,
                    hexstring_to_float(color[3:5])/255,
                    hexstring_to_float(color[5:7])/255]
            if n == 9:
                rgba.append(hexstring_to_float(color[7:9])/255)
            else:
                rgba.append(1.0)
        except ValueError:
            rgba = None
    return rgba

class dictproperty(object):
    """ Property implementation for dictionaries """

    class _proxy(object):

        def __init__(self, obj, fget, fset, fdel):
            self._obj = obj
            self._fget = fget
            self._fset = fset
            self._fdel = fdel

        def __getitem__(self, key):
            if self._fget is None:
                raise TypeError("can't read item")
            return self._fget(self._obj, key)

        def __setitem__(self, key, value):
            if self._fset is None:
                raise TypeError("can't set item")
            self._fset(self._obj, key, value)

        def __delitem__(self, key):
            if self._fdel is None:
                raise TypeError("can't delete item")
            self._fdel(self._obj, key)

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self._fget = fget
        self._fset = fset
        self._fdel = fdel
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return self._proxy(obj, self._fget, self._fset, self._fdel)

def unpack_name_value_list(_list, num_values=2, key_type = str):
    """
    Converts a list of strings into a dict of tuples.
    Sample list: ['LWIN:label:super', ...]
    ":" in a value must be escaped as "\:"
    "\" in a value must be escaped as "\\"
    """
    result = {}

    # Awkward fixed regexes; todo: Allow arbirary number of values
    if num_values == 1:
        pattern = re.compile(r"""([^\s:]+)             # name
                                 : ((?:\\.|[^\\:])*)   # first value
                             """, re.VERBOSE)
    elif num_values == 2:
        pattern = re.compile(r"""([^\s:]+)             # name
                                 : ((?:\\.|[^\\:])*)   # first value
                                 : ((?:\\.|[^\\:])*)   # second value
                             """, re.VERBOSE)
    else:
        assert(False)  # unsupported number of values

    for text in _list:
        tuples = pattern.findall(text)
        if tuples:
            a = []
            for t in tuples[0]:
                t = t.replace("\\\\", "\\")   # unescape backslash
                t = t.replace("\\:", ":")     # unescape separator
                a.append(t)

            if key_type == str:
                item = {a[0] : (a[1:])}
            elif key_type == int:
                item = {int(a[0]) : (a[1:])}
            else:
                assert(False)
            result.update(item)

    return result

def pack_name_value_list(tuples, field_sep=":", name_sep=":"):
    """
    Converts a dict of tuples to a string array. It creates one string
    per dict key, with the key-string separated by <name_sep> and
    individual tuple elements separated by <field_sep>.
    """
    result = []
    for t in list(tuples.items()):
        text = str(t[0])
        sep = name_sep
        for value in t[1]:
            value = value.replace("\\", "\\\\")   # escape backslash
            value = value.replace(sep, "\\"+sep)  # escape separator
            text += sep + '%s' % value
            sep = field_sep
        result.append(text)
    return result

def merge_tuple_strings(text1, text2):
    """
    Existing entries in text1 will be kept or overwritten by text2.
    """
    tuples1 = unpack_name_value_tuples(text1)
    tuples2 = unpack_name_value_tuples(text2)
    for key,values in list(tuples2.items()):
        tuples1[key] = values
    return pack_name_value_tuples(tuples1)


class Rect:
    """
    Simple rectangle class.
    Left and top are included, right and bottom excluded.
    Attributes can be accessed by name or by index, e.g. rect.x or rect[0].
    """

    attributes = ("x", "y", "w", "h")

    def __init__(self, x = 0, y = 0, w = 0, h = 0):
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def __len__(self):
        return 4

    def __getitem__(self, index):
        """ Collection interface for rvalues, unpacking with '*' operator """
        return getattr(self, self.attributes[index])

    def __setitem__(self, index, value):
        """ Collection interface for lvalues """
        return setattr(self, self.attributes[index], value)

    def __str__(self):
        return "Rect(" + \
            " ".join("{}={:.1f}".format(a, getattr(self, a)) \
                     for a in self.attributes) + \
            ")"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.x == other.x and \
               self.y == other.y and \
               self.w == other.w and \
               self.h == other.h

    def __ne__(self, other):
        return self.x != other.x or \
               self.y != other.y or \
               self.w != other.w or \
               self.h != other.h

    @staticmethod
    def from_extents(x0, y0, x1, y1):
        """
        New Rect from two points.
        x0 and y0 are considered inside, x1 and y1 are just outside the Rect.
        """
        return Rect(x0, y0, x1 - x0, y1 - y0)

    @staticmethod
    def from_position_size(position, size):
        """
        New Rect from two tuples.
        """
        return Rect(position[0], position[1], size[0], size[1])

    @staticmethod
    def from_points(p0, p1):
        """
        New Rect from two points, left-top and right-botton.
        The former lies inside, while the latter is considered to be
        just outside the rect.
        """
        return Rect(p0[0], p0[1], p1[0] - p0[0], p1[1] - p0[1])

    def to_extents(self):
        return self.x, self.y , self.x + self.w, self.y + self.h

    def to_position_size(self):
        return (self.x, self.y), (self.w, self.h)

    def copy(self):
        return Rect(self.x, self.y, self.w, self.h)

    def is_empty(self):
        return self.w <= 0 or self.h <= 0

    def get_position(self):
        return (self.x, self.y)

    def get_size(self):
        return (self.w, self.h)

    def get_center(self):
        return (self.x + self.w / 2.0, self.y + self.h / 2.0)

    def top(self):
        return self.y

    def left(self):
        return self.x

    def right(self):
        return self.x + self.w

    def bottom(self):
        return self.y + self.h

    def left_top(self):
        return self.x, self.y

    def is_point_within(self, point):
        """ True, if the given point lies inside the rectangle """
        if self.x <= point[0] and \
           self.x + self.w > point[0] and \
           self.y <= point[1] and \
           self.y + self.h > point[1]:
            return True
        return False

    def round(self):
        return Rect(round(self.x), round(self.y), round(self.w), round(self.h))

    def int(self):
        return Rect(int(self.x), int(self.y), int(self.w), int(self.h))

    def scale(self, kx, ky = None):
        if ky == None:
            ky = kx
        return Rect(self.x * kx, self.y * ky, self.w * kx, self.h * ky)

    def offset(self, dx, dy):
        """
        Returns a new Rect displaced by dx and dy.
        """
        return Rect(self.x + dx, self.y + dy, self.w, self.h)

    def inflate(self, dx, dy = None):
        """
        Returns a new Rect which is larger by dx and dy on all sides.
        """
        if dy is None:
            dy = dx
        return Rect(self.x-dx, self.y-dy, self.w+2*dx, self.h+2*dy)

    def apply_border(self, left, top, right, bottom):
        """
        Returns a new Rect which is larger by l, t, r, b on all sides.
        """
        return Rect(self.x-left, self.y-top,
                    self.w+left+right, self.h+top+bottom)

    def deflate(self, dx, dy = None):
        """
        Returns a new Rect which is smaller by dx and dy on all sides.
        """
        if dy is None:
            dy = dx
        return Rect(self.x+dx, self.y+dy, self.w-2*dx, self.h-2*dy)

    def grow(self, kx, ky = None):
        """
        Returns a new Rect with its size multiplied by kx, ky.
        """
        if ky is None:
            ky = kx
        w = self.w * kx
        h = self.h * ky
        return Rect(self.x + (self.w - w) / 2.0,
                    self.y + (self.h - h) / 2.0,
                    w, h)

    def intersects(self, rect):
        """
        Doctests:
        >>> Rect(0, 0, 1, 1).intersects(Rect(0, 0, 1, 1))
        True
        >>> Rect(0, 0, 1, 1).intersects(Rect(1, 0, 1, 1))
        False
        >>> Rect(1, 0, 1, 1).intersects(Rect(0, 0, 1, 1))
        False
        >>> Rect(0, 0, 1, 1).intersects(Rect(0, 1, 1, 1))
        False
        >>> Rect(0, 1, 1, 1).intersects(Rect(0, 0, 1, 1))
        False
        """
        #return not self.intersection(rect).is_empty()
        return not (self.x >= rect.x + rect.w or \
                    self.x + self.w <= rect.x or \
                    self.y >= rect.y + rect.h or \
                    self.y + self.h <= rect.y)

    def intersection(self, rect):
       x0 = max(self.x, rect.x)
       y0 = max(self.y, rect.y)
       x1 = min(self.x + self.w,  rect.x + rect.w)
       y1 = min(self.y + self.h,  rect.y + rect.h)
       if x0 > x1 or y0 > y1:
           return Rect()
       else:
           return Rect(x0, y0, x1 - x0, y1 - y0)

    def union(self, rect):
       x0 = min(self.x, rect.x)
       y0 = min(self.y, rect.y)
       x1 = max(self.x + self.w,  rect.x + rect.w)
       y1 = max(self.y + self.h,  rect.y + rect.h)
       return Rect(x0, y0, x1 - x0, y1 - y0)

    def inscribe_with_aspect(self, rect, x_align = 0.5, y_align = 0.5):
        """ Returns a new Rect with the aspect ratio of self
            that fits inside the given rectangle.
        """
        if self.is_empty() or rect.is_empty():
            return Rect()

        src_aspect = self.w / float(self.h)
        dst_aspect = rect.w / float(rect.h)

        result = rect.copy()
        if dst_aspect > src_aspect:
            result.w = rect.h * src_aspect
            result.x += x_align * (rect.w - result.w)
        else:
            result.h = rect.w / src_aspect
            result.y += y_align * (rect.h - result.h)
        return result

    def resize_to_aspect(self, aspect_rect):
        """
        Resize self to get the aspect ratio of aspect_rect.
        """
        if self.is_empty() or aspect_rect.is_empty():
            return Rect()

        src_aspect = aspect_rect.w / float(aspect_rect.h)
        dst_aspect = self.w / float(self.h)

        result = self.copy()
        if dst_aspect > src_aspect:
            result.w = self.h * src_aspect
        else:
            result.h = self.w / src_aspect
        return result

    def resize_to_aspect_range(self, aspect_rect, aspect_change_range):
        """
        Resize self to get the aspect ratio of aspect_rect, but limited
        but the given aspect range.
        """
        if self.is_empty() or aspect_rect.is_empty():
            return Rect()

        r = aspect_rect
        if r.h:
            a0 = r.w / float(r.h)
            a0_max = a0 * aspect_change_range[1]
            a1 = self.w / float(self.h)
            a = min(a1, a0_max)

            r = Rect(0, 0, a, 1.0)
            r = Rect(0, 0, a, 1.0)
        return self.resize_to_aspect(r)

    def align_rect(self, rect, x_align = 0.5, y_align = 0.5):
        """
        Aligns the given rect inside of self.
        x/y_align = 0.5 centers rect.
        """
        x = self.x + (self.w - rect.w) * x_align
        y = self.y + (self.h - rect.h) * y_align
        return Rect(x, y, rect.w, rect.h)

    def align_at_point(self, x, y, x_align = 0.5, y_align = 0.5):
        """
        Aligns the given rect to a point.
        x/y_align = 0.5 centers rect.
        """
        x = x - self.w * x_align
        y = y - self.h * y_align
        return Rect(x, y, self.w, self.h)

    def subdivide(self, columns, rows, x_spacing = None, y_spacing = None):
        """ Divide self into columns x rows sub-rectangles """
        if y_spacing is None:
            y_spacing = x_spacing
        if x_spacing is None:
            x_spacing = 0

        x, y, w, h = self
        ws = (self.w - (columns - 1) * x_spacing) / float(columns)
        hs = (self.h - (rows - 1)    * y_spacing) / float(rows)

        rects = []
        y = self.y
        for row in range(rows):
            x = self.x
            for column in range(columns):
                rects.append(Rect(x, y, ws, hs))
                x += ws + x_spacing
            y += hs + y_spacing

        return rects


def brighten(amount, r, g, b, a=0.0):
    """ Make the given color brighter by amount a [-1.0...1.0] """
    h, l, s = colorsys.rgb_to_hls(r, g, b)
    l += amount
    if l > 1.0:
        l = 1.0
    if l < 0.0:
        l = 0.0
    return list(colorsys.hls_to_rgb(h, l, s)) + [a]

def linint_rgba(k, rgba1, rgba2):
    """ interpolate between two colors """
    linint = lambda k, a, b: a + (b - a) * k
    return [linint(k, rgba1[0], rgba12[0]),
            linint(k, rgba1[1], rgba12[1]),
            linint(k, rgba1[2], rgba12[2]),
            linint(k, rgba1[3], rgba12[3])]

def roundrect_arc(context, rect, r = 15):
    x0,y0 = rect.x, rect.y
    x1,y1 = x0 + rect.w, y0 + rect.h

    # top left
    context.move_to(x0+r, y0)

    # top right
    context.line_to(x1-r,y0)
    context.arc(x1-r, y0+r, r, -pi/2, 0)

    # bottom right
    context.line_to(x1, y1-r)
    context.arc(x1-r, y1-r, r, 0, pi/2)

    # bottom left
    context.line_to(x0+r, y1)
    context.arc(x0+r, y1-r, r, pi/2, pi)

    # top left
    context.line_to(x0, y0+r)
    context.arc(x0+r, y0+r, r, pi, pi*1.5)

    context.close_path ()

def roundrect_curve(context, rect, r_pct = 100):
    """
    Uses B-splines for less even looks than with arcs, but
    still allows for approximate circles at r_pct = 100.
    """
    x0 = rect.x
    y0 = rect.y
    w  = rect.w
    h  = rect.h
    x1 = x0 + w
    y1 = y0 + h

    r = min(w, h) * min(r_pct/100.0, 0.5) # full range at 50%
    k = (r-1) * r_pct/200.0 # position of control points for circular curves

    line_to = context.line_to
    curve_to = context.curve_to

    # top left
    context.move_to(x0+r, y0)

    # top right
    line_to(x1-r,y0)
    curve_to(x1-k, y0, x1, y0+k, x1, y0+r)

    # bottom right
    line_to(x1, y1-r)
    curve_to(x1, y1-k, x1-k, y1, x1-r, y1)

    # bottom left
    line_to(x0+r, y1)
    curve_to(x0+k, y1, x0, y1-k, x0, y1-r)

    # top left
    line_to(x0, y0+r)
    curve_to(x0, y0+k, x0+k, y0, x0+r, y0)

    context.close_path ()

def rounded_polygon(cr, coords, r_pct, chamfer_size):
    path = polygon_to_rounded_path(coords, r_pct, chamfer_size)
    rounded_polygon_path_to_cairo_path(cr, path)

def polygon_to_rounded_path(coords, r_pct, chamfer_size):
    """
    Doctests:
    # simple rectangle, chamfer radius 0.
    >>> coords = [0, 0, 10, 0, 10, 10, 0, 10]
    >>> polygon_to_rounded_path(coords, 0, 0) # doctest: +NORMALIZE_WHITESPACE
    [(0.0, 0.0), (10.0, 0.0), (10.0, 0.0, 10.0, 0.0, 10.0, 0.0),
     (10.0, 10.0), (10.0, 10.0, 10.0, 10.0, 10.0, 10.0),
     (0.0, 10.0), (0.0, 10.0, 0.0, 10.0, 0.0, 10.0),
     (0.0, 0.0), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)]
    """
    path = []
    r = chamfer_size * 2.0 * min(r_pct/100.0, 0.5) # full range at 50%

    n = len(coords)
    for i in range(0, n, 2):
        i0 = i
        i1 = i + 2
        if i1 >= n:
            i1 -= n
        i2 = i + 4
        if i2 >= n:
            i2 -= n
        x0 = coords[i0]
        y0 = coords[i0+1]
        x1 = coords[i1]
        y1 = coords[i1+1]
        x2 = coords[i2]
        y2 = coords[i2+1]

        vax = x1 - x0
        vay = y1 - y0
        la = sqrt(vax*vax + vay*vay)
        uax = vax / la
        uay = vay / la

        vbx = x2 - x1
        vby = y2 - y1
        lb = sqrt(vbx*vbx + vby*vby)
        ubx = vbx / lb
        uby = vby / lb

        ra = min(r, la * 0.5)     # offset of curve begin and end
        rb = min(r, lb * 0.5)
        ka = (ra-1) * r_pct/200.0 # offset of control points
        kb = (rb-1) * r_pct/200.0

        if i == 0:
            x = x0 + ra*uax
            y = y0 + ra*uay
            path.append((x, y))

        x = x1 - ra*uax
        y = y1 - ra*uay
        path.append((x, y))

        x = x1 + rb*ubx
        y = y1 + rb*uby
        c0x = x1 - ka*uax
        c0y = y1 - ka*uay
        c1x = x1 + kb*ubx
        c1y = y1 + kb*uby
        path.append((x, y, c0x, c0y, c1x, c1y))

    return path

def rounded_polygon_path_to_cairo_path(cr, path):
    if path:
        cr.move_to(*path[0])
        for i in range(1, len(path), 2):
            p = path[i]
            cr.line_to(p[0], p[1])
            p = path[i+1]
            cr.curve_to(p[2], p[3], p[4], p[5], p[0], p[1])
        cr.close_path()

def rounded_path(cr, path, r_pct, chamfer_size):
    for polygon in path.iter_polygons():
        rounded_polygon(cr, polygon, r_pct, chamfer_size)

def round_corners(cr, r, x, y, w, h):
    """
    Paint 4 round corners.
    Currently x, y are ignored and assumed to be 0.
    """
    # top-left
    cr.curve_to (0, r, 0, 0, r, 0)
    cr.line_to (0, 0)
    cr.close_path()
    cr.fill()
    # top-right
    cr.curve_to (w, r, w, 0, w - r, 0)
    cr.line_to (w, 0)
    cr.close_path()
    cr.fill()
    # bottom-left
    cr.curve_to (r, h, 0, h, 0, h - r)
    cr.line_to (0, h)
    cr.close_path()
    cr.fill()
    # bottom-right
    cr.curve_to (w, h - r, w, h, w - r, h)
    cr.line_to (w, h)
    cr.close_path()
    cr.fill()

def gradient_line(rect, alpha):
    # Find rotated gradient start and end points.
    # Line end points follow the largest extent of the rotated rectangle.
    # The gradient reaches across the entire rectangle.
    x0, y0, w, h = rect.x, rect.y, rect.w, rect.h
    a = w / 2.0
    b = h / 2.0
    coords = [(-a, -b), (a, -b), (a, b), (-a, b)]
    vx = [c[0]*cos(alpha)-c[1]*sin(alpha) for c in coords]
    dx = max(vx) - min(vx)
    r = dx / 2.0
    return (r * cos(alpha) + x0 + a,
            r * sin(alpha) + y0 + b,
           -r * cos(alpha) + x0 + a,
           -r * sin(alpha) + y0 + b)
import cairo
def drop_shadow(cr, pattern, bounds, blur_radius = 4.0, offset = (0, 0),
                                  alpha=0.06, steps=4):
    """
    Mostly works, but has issues with clipping artefacts for
    damage rects smaller than the full window rect.
    """
    origin = bounds.get_center()
    cr.set_source_rgba(0.0, 0.0, 0.0, alpha)
    for i in range(steps):

        x = (i if i else 0.5) / float(steps)
        k = sqrt(abs(log(1-x))) * 0.7 * blur_radius # gaussian
        #k = i / float(steps) * blur_radius         # linear

        x_scale = (bounds.w + k) / bounds.w
        y_scale = (bounds.h + k) / bounds.h
        cr.save()
        cr.translate(*origin)
        cr.scale(x_scale, y_scale)
        cr.translate(-origin[0] + offset[0], -origin[1] + offset[1])

        cr.mask(pattern)
        cr.restore()

@contextmanager
def timeit(s, out=sys.stdout):
    import time, gc

    if out:
        gc.collect()
        gc.collect()
        gc.collect()

        t = time.time()
        text = s if s else "timeit"
        out.write("%-15s " % text)
        out.flush()
        yield None
        out.write("%10.3fms\n" % ((time.time() - t)*1000))
    else:
        yield None

class Fade:
    """ Helper for opacity fading """
    @staticmethod
    def sin_fade(start_time, duration, start_value, target_value):
        elapsed = time.time() - start_time
        if duration:
            lin_progress = min(1.0, elapsed / duration)
        else:
            lin_progress = 1.0
        return(Fade.sin_int(lin_progress, start_value, target_value),
               lin_progress >= 1.0)

    @staticmethod
    def sin_int(lin_progress, start_value, target_value):
        sin_progress = (sin(lin_progress * pi - pi / 2.0) + 1.0) / 2.0
        return sin_progress * (target_value - start_value) + start_value


class TreeItem(object):
    """
    Abstract base class of tree nodes.
    Base class of nodes in  layout- and color scheme tree.
    """

    # id string of the item
    id = None

    # parent item in the tree
    parent = None

    # child items
    items = ()

    def set_items(self, items):
        self.items = items
        for item in items:
            item.parent = self

    def append_item(self, item):
        if self.items:
            self.items.append(item)
        else:
            self.items = [item]

        item.parent = self

    def append_items(self, items):
        if self.items:
            self.items += items
        else:
            self.items = items

        for item in items:
            item.parent = self

    def get_parent(self):
        return self.parent

    def find_ids(self, ids):
        """ find all items with matching id """
        for item in self.iter_items():
            if item.id in ids:
                yield item

    def find_classes(self, item_classes):
        """ find all items with matching id """
        for item in self.iter_items():
            if isinstance(item, item_classes):
                yield item

    def iter_items(self):
        """
        Iterates through all items of the tree.
        """
        yield self

        for item in self.items:
            for child in item.iter_items():
                yield child

    def iter_depth_first(self):
        """
        Iterates depth first through the tree.
        """
        for item in self.items:
            for child in item.iter_depth_first():
                yield child

        yield self

    def iter_to_root(self):
        item = self
        while item:
            yield item
            item = item.parent


class Version(object):
    """ Simple class to encapsulate a version number """
    major = 0
    minor = 0

    def __init__(self, major, minor = 0):
        self.major = major
        self.minor = minor

    def __str__(self):
        return self.to_string()

    @staticmethod
    def from_string(version):
        components = version.split(".")

        major = 0
        minor = 0
        try:
            if len(components) >= 1:
                major = int(components[0])
            if len(components) >= 2:
                minor = int(components[1])
        except ValueError:
            pass

        return Version(major, minor)

    def to_string(self):
        return "{major}.{minor}".format(major=self.major, minor=self.minor)

    def __eq__(self, other): return self._cmp(other) == 0
    def __ne__(self, other): return self._cmp(other) != 0
    def __lt__(self, other): return self._cmp(other) < 0
    def __le__(self, other): return self._cmp(other) <= 0
    def __gt__(self, other): return self._cmp(other) > 0
    def __ge__(self, other): return self._cmp(other) >= 0

    def _cmp(self, other):
        if self.major < other.major:
            return -1
        if self.major > other.major:
            return 1
        if self.minor < other.minor:
            return -1
        if self.minor > other.minor:
            return 1
        return 0


class Process:
    """ Process utilities """

    @staticmethod
    def get_cmdline(pid):
        """ Returns the command line for process id pid """
        cmdline = ""
        with open("/proc/%s/cmdline" % pid) as f:
            cmdline = f.read()
        return cmdline.split("\0")

    @staticmethod
    def get_process_name(pid):
        cmdline = Process.get_cmdline(pid)
        if cmdline:
            return os.path.basename(cmdline[0])
        return ""

    @staticmethod
    def get_launch_process_cmdline():
        """ Checks if this process was launched by <process_name> """
        ppid = os.getppid()
        if ppid:
            cmdline = Process.get_cmdline(ppid)
            return cmdline
        return []

    @staticmethod
    def was_launched_by(process_name):
        """ Checks if this process was launched by <process_name> """
        cmdline = " ".join(Process.get_launch_process_cmdline())
        return process_name in cmdline


def exists_in_path(basename):
    """
    Does a file with this basename exist anywhere in PATH's directories?
    """
    for path in os.environ["PATH"].split(os.pathsep):
        filename = os.path.join(path, basename)
        if os.path.isfile(filename):
            return True
    return False

def chmodtree(path, mode = 0o777, only_dirs = False):
    """
    Change permissions of all files of the given directory tree.
    Raises OSError.
    """
    os.chmod(path, mode)
    for root, dirs, files in os.walk(path):
        for d in dirs:
            os.chmod(os.path.join(root, d), mode)
        if not only_dirs:
            for f in files:
                os.chmod(os.path.join(root, f), mode)

def unicode_str(obj, encoding = "utf-8"):
    """
    Safe str() function that always returns an unicode string.
    Do nothing if the string was already unicode.
    """
    if sys.version_info.major >= 3:  # python 3?
        return str(obj)

    if type(obj) == unicode:         # unicode string?
        return obj

    if hasattr(obj, "__unicode__"):  # Exception object?
        return unicode(obj)

    return str(obj).decode("utf-8")  # strings, numbers, ...

def open_utf8(filename, mode = "r"):
    """
    Python 2 compatible replacement for builtin open().
    Python 3 added the encoding parameter.
    """
    if sys.version_info.major == 2:
        return open(filename, mode)
    else:
        return open(filename, mode=mode, encoding="UTF-8")

def permute_mask(mask):
    """
    Return all permutations of the bits in mask.

    Doctests:
    >>> permute_mask(1)
    [0, 1]
    >>> permute_mask(5)
    [0, 1, 4, 5]
    >>> permute_mask(14)
    [0, 2, 4, 6, 8, 10, 12, 14]
    """
    bit_masks = [bit_mask for bit_mask in (1<<bit for bit in range(8)) \
                 if mask & bit_mask]
    n = len(bit_masks)
    perms = []
    for i in range(2**n):
        m = 0
        for bit in range(n):
            if i & 1<<bit:
                m |= bit_masks[bit]
        perms.append(m)
    return perms


class Translation:
    """
    Translations occasionally contain errors in format fields that
    prevent onboard from starting up. This class aims to catch these
    errors gracefully, report and ignore bad translations and then
    just go on.

    Common errors have been:
        - bad field names, e.g. "{filename}" was translated "{path}"
        - bad anonymous fields, e.g. "{}" was translated "{ }"
    """
    @staticmethod
    def install(domain):
        """ setup gettext, install _() function for all modules """

        try:
            import builtins
        except ImportError:
            builtins = sys.modules["__builtin__"]  # python 2.x

        t = Translation()
        t.translation = gettext.translation(domain, fallback=True)

        builtins.__dict__['_'] = t.ugettext
        builtins.__dict__['_format'] = t.format

    def ugettext(self, msgid):
        if not msgid:
            return unicode_str("")
        if sys.version_info.major < 3:  # python 2?
            return self.translation.ugettext(msgid)
        return self.translation.gettext(msgid)

    def format(self, msgid, *args, **kwargs):
        """ Safe replacement for str.format() """
        msgstr = self.ugettext(msgid)
        try:
            result = msgstr.format(*args, **kwargs)
        except (KeyError, IndexError, UnicodeDecodeError) as e:
            result = msgid.format(*args, **kwargs)

            _logger.warning("_format: Skipping bad translation "
                            "msgid='{}' msgstr='{}' {}: {}" \
                            .format(msgid, msgstr,
                                    e.__class__.__name__,
                                    unicode_str(e)))
        return result


class EventSource(object):
    """ Simple event handling based on python callbacks """
    _event_queue = None  # for optional async delivery

    def __init__(self, event_names):
        self._callbacks = dict((e,[]) for e in event_names)

    def cleanup(self):
        self.flush_events()

    def connect(self, event_name, callback):
        callbacks = self._callbacks[event_name]
        if not callback in callbacks:
            callbacks.append(callback)

    def disconnect(self, event_name, callback):
        callbacks = self._callbacks[event_name]
        if callback in callbacks:
            callbacks.remove(callback)

    def has_listeners(self, event_names = None):
        """
        Are there callbacks registered for the given event_names or any event?
        """
        if event_names:
            return any(bool(self._callbacks[name]) for name in event_names)
        return any(bool(value) for value in self._callbacks.values())

    def emit(self, event_name, *args, **kwargs):
        """
        Send event, call all listener's callbacks.
        """
        #print("emit", event_name, list(args), kwargs)
        for callback in self._callbacks[event_name]:
            callback(*args, **kwargs)

    def emit_async(self, event_name, *args, **kwargs):
        """
        Queue up asynchronous event.
        """
        #print("emit_async", event_name, list(args), kwargs)
        event = (event_name, args, kwargs)
        if self._event_queue is None:
            self._event_queue = [event]
            GLib.idle_add(self.flush_events)
        else:
            self._event_queue.append(event)

    def flush_events(self):
        """
        Send pending asynchronous events.
        """
        if not self._event_queue is None:
            for event_name, args, kwargs in self._event_queue:
                self.emit(event_name, *args, **kwargs)
            self.clear_events()

    def clear_events(self):
        """
        Cancel pending asynchronous events.
        """
        self._event_queue = None


class XDGDirs:
    """
    Build paths compliant with XDG Base Directory Specification.
    http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html

    Doctests:

    >>> old_env = os.environ.copy()
    >>> os.environ["HOME"] = "/home/test_user"

    # XDG_CONFIG_HOME unavailable
    >>> os.environ["XDG_CONFIG_HOME"] = ""
    >>> XDGDirs.get_config_home("onboard/test.dat")
    '/home/test_user/.config/onboard/test.dat'

    # XDG_CONFIG_HOME available
    >>> os.environ["XDG_CONFIG_HOME"] = "/home/test_user/.config_home"
    >>> XDGDirs.get_config_home("onboard/test.dat")
    '/home/test_user/.config_home/onboard/test.dat'

    # XDG_DATA_HOME unavailable
    >>> os.environ["XDG_DATA_HOME"] = ""
    >>> XDGDirs.get_data_home("onboard/test.dat")
    '/home/test_user/.local/share/onboard/test.dat'

    # XDG_DATA_HOME available
    >>> os.environ["XDG_DATA_HOME"] = "/home/test_user/.data_home"
    >>> XDGDirs.get_data_home("onboard/test.dat")
    '/home/test_user/.data_home/onboard/test.dat'

    # XDG_CONFIG_DIRS unvailable
    >>> os.environ["XDG_CONFIG_HOME"] = ""
    >>> os.environ["XDG_CONFIG_DIRS"] = ""
    >>> XDGDirs.get_all_config_dirs("onboard/test.dat")
    ['/home/test_user/.config/onboard/test.dat', '/etc/xdg/onboard/test.dat']

    # XDG_CONFIG_DIRS available
    >>> os.environ["XDG_CONFIG_HOME"] = ""
    >>> os.environ["XDG_CONFIG_DIRS"] = "/etc/xdg/xdg-ubuntu:/etc/xdg"
    >>> XDGDirs.get_all_config_dirs("onboard/test.dat")
    ['/home/test_user/.config/onboard/test.dat', \
'/etc/xdg/xdg-ubuntu/onboard/test.dat', \
'/etc/xdg/onboard/test.dat']

    # XDG_DATA_DIRS unvailable
    >>> os.environ["XDG_DATA_HOME"] = ""
    >>> os.environ["XDG_DATA_DIRS"] = ""
    >>> XDGDirs.get_all_data_dirs("onboard/test.dat")
    ['/home/test_user/.local/share/onboard/test.dat', \
'/usr/local/share/onboard/test.dat', \
'/usr/share/onboard/test.dat']

    # XDG_DATA_DIRS available
    >>> os.environ["XDG_DATA_HOME"] = ""
    >>> os.environ["XDG_DATA_DIRS"] = "/usr/share/gnome:/usr/local/share/:/usr/share/"
    >>> XDGDirs.get_all_data_dirs("onboard/test.dat")
    ['/home/test_user/.local/share/onboard/test.dat', \
'/usr/share/gnome/onboard/test.dat', \
'/usr/local/share/onboard/test.dat', \
'/usr/share/onboard/test.dat']

    >>> os.environ = old_env
    """

    @staticmethod
    def get_config_home(file = None):
        """
        User specific config directory.
        """
        path = os.environ.get("XDG_CONFIG_HOME")
        if path and not os.path.isabs(path):
            _logger.warning("XDG_CONFIG_HOME doesn't contain an absolute path,"
                            "ignoring.")
            path = None
        if not path:
            path = os.path.join(os.path.expanduser("~"), ".config")

        if file:
            path = os.path.join(path, file)

        return path

    @staticmethod
    def get_config_dirs():
        """
        Config directories ordered by preference.
        """
        paths = []

        value = os.environ.get("XDG_CONFIG_DIRS")
        if not value:
            value = "/etc/xdg"

        paths = value.split(":")
        paths = [p for p in paths if os.path.isabs(p)]

        return paths

    @staticmethod
    def get_all_config_dirs(file = None):
        paths = [XDGDirs.get_config_home()] + XDGDirs.get_config_dirs()

        if file:
            paths = [os.path.join(p, file) for p in paths]

        return paths

    @staticmethod
    def find_config_files(file):
        """ Find file in all config directories, highest priority first. """
        paths = XDGDirs.get_all_config_dirs(file)
        return [p for p in paths if os.path.isfile(path) and \
                                    os.access(filename, os.R_OK)]

    @staticmethod
    def find_config_file(file):
        """ Find file of highest priority """
        paths = XDGDirs.find_config_files(file)
        if paths:
            return paths[0]
        return None

    @staticmethod
    def get_data_home(file = None):
        """
        User specific data directory.
        """
        path = os.environ.get("XDG_DATA_HOME")
        if path and not os.path.isabs(path):
            _logger.warning("XDG_DATA_HOME doesn't contain an absolute path,"
                            "ignoring.")
            path = None
        if not path:
            path = os.path.join(os.path.expanduser("~"), ".local", "share")

        if file:
            path = os.path.join(path, file)

        return path

    @staticmethod
    def get_data_dirs():
        """
        Data directories ordered by preference.
        """
        paths = []

        value = os.environ.get("XDG_DATA_DIRS")
        if not value:
            value = "/usr/local/share/:/usr/share/"

        paths = value.split(":")
        paths = [p for p in paths if os.path.isabs(p)]

        return paths

    @staticmethod
    def get_all_data_dirs(file = None):
        paths = [XDGDirs.get_data_home()] + XDGDirs.get_data_dirs()

        if file:
            paths = [os.path.join(p, file) for p in paths]

        return paths

    @staticmethod
    def find_data_files(file):
        """ Find file in all data directories, highest priority first. """
        paths = XDGDirs.get_all_data_dirs(file)
        return [p for p in paths if os.path.isfile(path) and \
                                    os.access(filename, os.R_OK)]

    @staticmethod
    def find_data_file(file):
        """ Find file of highest priority """
        paths = XDGDirs.find_data_files(file)
        if paths:
            return paths[0]
        return None

    def assure_user_dir_exists(path):
        """
        If necessary create user XDG directory.
        Raises OSError.
        """
        exists = os.path.exists(path)
        if not exists:
            try:
                os.makedirs(path, mode = 0o700)
                exists = True
            except OSError as ex:
                _logger.error(_format("failed to create directory '{}': {}",
                                       path, unicode_str(ex)))
                raise ex
        return exists


_tag_pattern = re.compile(
    """(?:
            <[\w\-_]+                         # tag
            (?:\s+[\w\-_]+=["'][^"']*["'])*  # attributes
            /?>
        ) |
        (?:
            </?[\w\-_]+>
        )
    """, re.UNICODE|re.DOTALL|re.VERBOSE)

def _iter_markup(markup):
    """
    Iterate over tag and non-tag sections of a markup string.

    Doctests:
    # Never yield for empty string
    >>> list(_iter_markup(""))
    []

    # must return tag- as well as non-tag sections
    >>> list(_iter_markup("<tt>test</tt>test2"))
    [('<tt>', True), ('test', False), ('</tt>', True), ('test2', False)]

    # should recognize tags with attributes
    >>> list(_iter_markup('<tag attr="value">'))
    [('<tag attr="value">', True)]
    >>> list(_iter_markup('<tag attr="v alue" attr2="234">'))
    [('<tag attr="v alue" attr2="234">', True)]

    # should recognize tags with end shortcut
    >>> list(_iter_markup('<tag/> <tag2 attr="value"/>'))
    [('<tag/>', True), (' ', False), ('<tag2 attr="value"/>', True)]

    # must not modify input, i.e. concatenated result must equal input text
    >>> markup = "asd <tt>t est\\n ds</tt> te st2 "
    >>> "".join([text for text, tag in _iter_markup(markup)]) == markup
    True
    """
    pos = 0
    matches = _tag_pattern.finditer(markup)
    for m in matches:
        text = markup[pos:m.start()]
        if text:
            yield text, False
        yield m.group(), True
        pos = m.end()

    text = markup[pos:]
    if text:
        yield text, False


def escape_markup(markup, preserve_tags = False):
    """
    Escape strings of uncertain content for use in Gtk markup.
    If requested, markup tags are skipped and won't be escaped.

    Doctests:
    >>> escape_markup("&<>")
    '&amp;&lt;&gt;'

    # tags must be escaped when preserve_tags is False
    >>> escape_markup("<big>&<></big>")
    '&lt;big&gt;&amp;&lt;&gt;&lt;/big&gt;'

    # tags must not be escaped when preserve_tags is True
    >>> escape_markup('<big>&</big>><tag attr="value"><</tag>1<3', True)
    '<big>&amp;</big>&gt;<tag attr="value">&lt;</tag>1&lt;3'

    # whitespace must be preserved
    >>> escape_markup("test <big> test2& </big> test3", True)
    'test <big> test2&amp; </big> test3'
    """
    result = ""
    for text, is_tag in _iter_markup(markup):
        if is_tag and preserve_tags:
            result += text
        else:
            try:
                result += GLib.markup_escape_text(text)
            except Exception as ex: # private exception gi._glib.GError
                _logger.error("markup_escape_text failed for "
                                "'{}': {}" \
                            .format(text, unicode_str(ex)))
    return result


class TermColors(object):
    """ Singleton class providing ANSI terminal color codes """

    # sequence ids
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, \
    BOLD, RESET = range(10)

    # sequence cache
    sequences = {}

    def __new__(cls, *args, **kwargs):
        """ Singleton magic.  """
        if not hasattr(cls, "self"):
            cls.self = object.__new__(cls, *args, **kwargs)
            cls.self.construct()
        return cls.self

    def __init__(self):
        """ Called multiple times, do not use.  """
        pass

    def construct(self):
        """ Singleton constructor, runs only once.  """

    def get(self, seq_id):
        """
        Return ANSI character sequence for the given sequence id,
        e.g. color index.
        """
        seq = self.sequences.get(seq_id)
        if seq is None:
            seq = ""
            if not seq_id is None:
                if seq_id >= self.BLACK and seq_id <= self.WHITE:
                    seq = self._tput("setaf " + str(seq_id))
                elif seq_id == self.BOLD:
                    seq = self._tput("bold")
                elif seq_id == self.RESET:
                    seq = self._tput("sgr0")
            self.sequences[seq_id] = seq
        return seq

    @staticmethod
    def _tput(params):
        try:
            s = subprocess.check_output(("tput " + params).split())
            return s.decode("ASCII")
        except subprocess.CalledProcessError:
            return ""






# -*- coding: utf-8 -*-

# Copyright © 2014 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Hide the keyboard on incoming physical keyboard events.
"""

from __future__ import division, print_function, unicode_literals


from Onboard.utils         import EventSource
from Onboard.XInput        import XIDeviceManager, XIEventType, XIEventMask

### Logging ###
import logging
_logger = logging.getLogger("AutoHide")
###############

from Onboard.Config import Config
config = Config()


class AutoHide(EventSource):
    """
    Hide Onboard when a physical keyboard is being used.
    """

    def __init__(self, keyboard):
        # There is only button-release to subscribe to currently,
        # as this is all CSButtonRemapper needs to detect the end of a click.
        EventSource.__init__(self, ["button-release"])

        self._keyboard = keyboard
        self._device_manager = None
        self._keyboard_slave_devices = None

    def cleanup(self):
        self._register_xinput_events(False)

    def is_enabled(self):
        return self._device_manager is not None

    def enable(self, enable, use_gtk=False):
        self.register_input_events(enable, use_gtk)

    def register_input_events(self, register, use_gtk=False):
        self._register_xinput_events(False)
        if register:
            if not use_gtk:  # can't do this with gtk yet
                if not self._register_xinput_events(True):
                    _logger.warning(
                        "XInput event source failed to initialize, "
                        "falling back to GTK.")

    def _register_xinput_events(self, register):
        """ Setup XInput event handling """
        success = True

        if register:
            self._device_manager = XIDeviceManager()
            if self._device_manager.is_valid():
                self._device_manager.connect("device-event",
                                             self._on_device_event)
                self._device_manager.connect("device-grab",
                                             self._on_device_grab)
                self._select_xinput_devices()
            else:
                success = False
                self._device_manager = None
        else:

            if self._device_manager:
                self._device_manager.disconnect("device-event",
                                                self._on_device_event)
                self._device_manager.disconnect("device-grab",
                                                self._on_device_grab)
                self._unselect_xinput_devices()
                self._device_manager = None

        return success

    def _select_xinput_devices(self):
        """ Select keyboard devices and the events we want to listen to. """

        self._unselect_xinput_devices()

        event_mask = XIEventMask.KeyPressMask | \
                     XIEventMask.KeyReleaseMask

        devices = self._device_manager.get_client_keyboard_attached_slaves()
        _logger.info("listening to keyboard devices: {}"
                     .format([(d.name, d.id, d.get_config_string())
                              for d in devices]))
        for device in devices:
            try:
                self._device_manager.select_events(None, device, event_mask)
            except Exception as ex:
                _logger.warning("Failed to select events for device "
                                "{id}: {ex}"
                                .format(id=device.id, ex=ex))
        self._keyboard_slave_devices = devices

    def _unselect_xinput_devices(self):
        if self._keyboard_slave_devices:
            for device in self._keyboard_slave_devices:
                try:
                    self._device_manager.unselect_events(None, device)
                except Exception as ex:
                    _logger.warning("Failed to unselect events for device "
                                    "{id}: {ex}"
                                    .format(id=device.id, ex=ex))
            self._keyboard_slave_devices = None

    def _on_device_grab(self, device, event):
        """ Someone grabbed/relased a device. Update our device list. """
        self._select_xinput_devices()

    def _on_device_event(self, event):
        """
        Handler for XI2 events.
        """
        event_type = event.xi_type

        # re-select devices on changes to the device hierarchy
        if event_type in XIEventType.HierarchyEvents or \
           event_type == XIEventType.DeviceChanged:
            self._select_xinput_devices()
            return

        if event_type == XIEventType.KeyPress or \
           event_type == XIEventType.KeyRelease:

            if not self._keyboard.is_auto_show_paused():
                if _logger.isEnabledFor(logging.INFO):
                    device = event.get_source_device()
                    device_name = device.name if device else "None"
                    _logger.info("Hiding keyboard and pausing "
                                "auto-show due to physical key-{} "
                                "{} from device '{}' ({})"
                                .format("press"
                                        if event_type == XIEventType.KeyPress
                                        else "release",
                                        event.keyval,
                                        device_name,
                                        event.source_id))

                if self._keyboard.is_visible():
                    if config.are_word_suggestions_enabled():
                        self._keyboard.discard_changes()

                    self._keyboard.set_visible(False)

            duration = config.auto_show.hide_on_key_press_pause
            if duration:
                if duration < 0.0:  # negative means auto-hide is off
                    duration = None
                self._keyboard.pause_auto_show(duration)

            return





# -*- coding: UTF-8 -*-

# Copyright © 2007 Martin Böhme <martin.bohm@kubuntu.org>
# Copyright © 2008-2009 Chris Jones <tortoise@tortuga>
# Copyright © 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
# Copyright © 2009, 2011-2014 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
KeyCommon hosts the abstract classes for the various types of Keys.
UI-specific keys should be defined in KeyGtk or KeyKDE files.
"""

from __future__ import division, print_function, unicode_literals

from math import pi
import re

from Onboard.utils import Rect, LABEL_MODIFIERS, Modifiers, \
                          polygon_to_rounded_path

from Onboard.Layout import LayoutItem

### Logging ###
import logging
_logger = logging.getLogger("KeyCommon")
###############

### Config Singleton ###
from Onboard.Config import Config
config = Config()
########################

(
    CHAR_TYPE,
    KEYSYM_TYPE,
    KEYCODE_TYPE,
    MACRO_TYPE,
    SCRIPT_TYPE,
    KEYPRESS_NAME_TYPE,
    BUTTON_TYPE,
    LEGACY_MODIFIER_TYPE,
    WORD_TYPE,
    CORRECTION_TYPE,
) = tuple(range(1, 11))

(
    SINGLE_STROKE_ACTION,  # press on button down, release on up (default)
    DELAYED_STROKE_ACTION, # press+release on button up (MENU)
    DOUBLE_STROKE_ACTION,  # press+release on button down and up, (CAPS, NMLK)
) = tuple(range(3))

actions = {
           "single-stroke"  : SINGLE_STROKE_ACTION,
           "delayed-stroke" : DELAYED_STROKE_ACTION,
           "double-stroke"  : DOUBLE_STROKE_ACTION,
          }

class StickyBehavior:
    """ enum for sticky key behaviors """
    (
        CYCLE,
        DOUBLE_CLICK,
        LATCH_ONLY,
        LOCK_ONLY,
        LATCH_LOCK_NOCYCLE,
        DOUBLE_CLICK_NOCYCLE,
        LATCH_NOCYCLE,
        LOCK_NOCYCLE,
        PUSH_BUTTON,
    ) = tuple(range(9))

    values = {"cycle"              : CYCLE,
              "dblclick"           : DOUBLE_CLICK,
              "latch"              : LATCH_ONLY,
              "lock"               : LOCK_ONLY,
              "latch-lock-nocycle" : LATCH_LOCK_NOCYCLE,
              "dblclick-nocycle"   : DOUBLE_CLICK_NOCYCLE,
              "latch-nocycle"      : LATCH_NOCYCLE,
              "lock-nocycle"       : LOCK_NOCYCLE,
              "push"               : PUSH_BUTTON,
             }

    @staticmethod
    def from_string(str_value):
        """ Raises KeyError """
        return StickyBehavior.values[str_value]

    @staticmethod
    def is_valid(behavior):
        return behavior in StickyBehavior.values.values()

    @staticmethod
    def can_latch(behavior):
        """
        Can sticky key enter latched state?
        Latched keys are automatically released when a
        non-sticky key is pressed.
        """
        return behavior in (StickyBehavior.CYCLE,
                            StickyBehavior.DOUBLE_CLICK,
                            StickyBehavior.LATCH_ONLY,
                            StickyBehavior.LATCH_LOCK_NOCYCLE,
                            StickyBehavior.DOUBLE_CLICK_NOCYCLE,
                            StickyBehavior.LATCH_NOCYCLE)

    @staticmethod
    def can_lock(behavior):
        return StickyBehavior.can_lock_on_single_click(behavior) or \
               StickyBehavior.can_lock_on_double_click(behavior)

    @staticmethod
    def can_lock_on_single_click(behavior):
        """
        Can sticky key enter locked state?
        Locked keys stay active until they are pressed again.
        """
        return behavior in (StickyBehavior.CYCLE,
                            StickyBehavior.LOCK_ONLY,
                            StickyBehavior.LATCH_LOCK_NOCYCLE,
                            StickyBehavior.LOCK_NOCYCLE)

    @staticmethod
    def can_lock_on_double_click(behavior):
        """
        Can sticky key enter locked state on double click?
        Locked keys stay active until they are pressed again.
        """
        return behavior == StickyBehavior.DOUBLE_CLICK or \
               behavior == StickyBehavior.DOUBLE_CLICK_NOCYCLE

    @staticmethod
    def can_cycle(behavior):
        """
        Can sticky key return to normal state?
        Latched keys are still automatically released when a
        non-sticky key is pressed.
        """
        return behavior in (StickyBehavior.CYCLE,
                            StickyBehavior.DOUBLE_CLICK,
                            StickyBehavior.LATCH_ONLY,
                            StickyBehavior.LOCK_ONLY)


class LOD:
    """ enum for level of detail """
    (
        MINIMAL,    # clearly visible reduced detail, fastest
        REDUCED,    # slightly reduced detail
        FULL,       # full detail
    ) = tuple(range(3))

class ImageSlot:
    NORMAL = 0
    ACTIVE = 1

class KeyCommon(LayoutItem):
    """
    library-independent key class. Specific rendering options
    are stored elsewhere.
    """

    # extended id for key specific theme tweaks
    # e.g. theme_id=DELE.numpad (with id=DELE)
    theme_id = None

    # extended id for layout specific tweaks
    # e.g. "hide.wordlist", for hide button in wordlist mode
    svg_id = None

    # optional id of a sublayout used as long-press popup
    popup_id = None

    # Type of action to do when key is pressed.
    action = None

    # Type of key stroke to send
    type = None

    # Data used in sending key strokes.
    code = None

    # Keys that stay stuck when pressed like modifiers.
    sticky = False

    # Behavior if sticky is enabled, see StickyBehavior.
    sticky_behavior = None

    # modifier bit
    modifier = None

    # True when key is being hovered over (not implemented yet)
    prelight = False

    # True when key is being pressed.
    pressed = False

    # True when key stays 'on'
    active = False

    # True when key is sticky and pressed twice.
    locked = False

    # True when Onboard is in scanning mode and key is highlighted
    scanned = False

    # True when action was triggered e.g. key-strokes were sent on press
    activated = False

    # Size to draw the label text in Pango units
    font_size = 1

    # Labels which are displayed by this key
    labels = None  # {modifier_mask : label, ...}

    # label that is currently displayed by this key
    label = ""

    # mod_mask for the currently configured label
    mod_mask = 0

    # smaller label of a currently invisible modifier level
    secondary_label = ""

    # Images displayed by this key (optional)
    image_filenames = None

    # horizontal label alignment
    label_x_align = config.DEFAULT_LABEL_X_ALIGN

    # vertical label alignment
    label_y_align = config.DEFAULT_LABEL_Y_ALIGN

    # label margin (x, y)
    label_margin = config.LABEL_MARGIN

    # tooltip text
    tooltip = None

    # can show label popup
    label_popup = True

###################

    def __init__(self):
        LayoutItem.__init__(self)

    def configure_label(self, mod_mask):
        SHIFT = Modifiers.SHIFT
        labels = self.labels

        if labels is None:
            self.label = self.secondary_label = ""
            return

        # primary label
        label = labels.get(mod_mask)
        if label is None:
            mask = mod_mask & LABEL_MODIFIERS
            label = labels.get(mask)

        # secondary label, usually the label of the shift state
        secondary_label = None
        if not label is None:
            if mod_mask & SHIFT:
                mask = mod_mask & ~SHIFT
            else:
                mask = mod_mask | SHIFT

            secondary_label = labels.get(mask)
            if secondary_label is None:
                mask = mask & LABEL_MODIFIERS
                secondary_label = labels.get(mask)

            # Only keep secondary labels that show different characters
            if not secondary_label is None and \
               secondary_label.upper() == label.upper():
                secondary_label = None

        if label is None:
            # legacy fallback for 0.98 behavior and virtkey until 0.61.0
            if mod_mask & Modifiers.SHIFT:
                if mod_mask & Modifiers.ALTGR and 129 in labels:
                    label = labels[129]
                elif 1 in labels:
                    label = labels[1]
                elif 2 in labels:
                    label = labels[2]

            elif mod_mask & Modifiers.ALTGR and 128 in labels:
                label = labels[128]

            elif mod_mask & Modifiers.CAPS:  # CAPS lock
                if 2 in labels:
                    label = labels[2]
                elif 1 in labels:
                    label = labels[1]

        if label is None:
            label = labels.get(0)

        if label is None:
            label = ""

        self.mod_mask = mod_mask
        self.label = label
        self.secondary_label = secondary_label

        # Don't let erroneous labels shrink their whole size group.
        self.ignore_group = label.startswith("0x")

    def draw_label(self, context = None):
        raise NotImplementedError()

    def set_labels(self, labels):
        self.labels = labels
        self.configure_label(0)

    def get_label(self):
        return self.label

    def get_secondary_label(self):
        return self.secondary_label

    def is_active(self):
        return not self.type is None

    def get_id(self):
        return ""

    def get_svg_id(self):
        return ""

    def set_id(self, id, theme_id = None, svg_id = None):
        self.theme_id, self.id = self.parse_id(id)
        if theme_id:
            self.theme_id = theme_id
        self.svg_id = self.id if not svg_id else svg_id

    @staticmethod
    def parse_id(value):
        """
        The theme id has the form <id>.<arbitrary identifier>, where
        the identifier should be a description of the location of
        the key relative to its surroundings, e.g. 'DELE.next-to-backspace'.
        Don't use layout names or layer ids for the theme id, they lose
        their meaning when layouts are copied or renamed by users.
        """
        theme_id = value
        id = value.split(".")[0]
        return theme_id, id

    @staticmethod
    def split_theme_id(theme_id):
        """
        Simple split in prefix (id) before the dot and suffix after the dot.
        """
        components = theme_id.split(".")
        if len(components) == 1:
            return components[0], ""
        return components[0], components[1]

    @staticmethod
    def build_theme_id(prefix, postfix):
        if postfix:
            return prefix + "." + postfix
        return prefix

    def get_similar_theme_id(self, prefix = None):
        if prefix is None:
            prefix = self.id
        theme_id = prefix
        comps = self.theme_id.split(".")[1:]
        if comps:
            theme_id += "." + comps[0]
        return theme_id

    def is_layer_button(self):
        return self.id.startswith("layer")

    def is_prediction_key(self):
        return self.id.startswith("prediction")

    def is_correction_key(self):
        return self.id.startswith("correction") or \
               self.id in ["expand-corrections"]

    def is_word_suggestion(self):
        return self.is_prediction_key() or self.is_correction_key()

    def is_modifier(self):
        """
        Modifiers are all latchable/lockable non-button keys:
        "LWIN", "RTSH", "LFSH", "RALT", "LALT",
        "RCTL", "LCTL", "CAPS", "NMLK"
        """
        return bool(self.modifier)

    def is_click_type_key(self):
        return self.id in ["singleclick",
                           "secondaryclick",
                           "middleclick",
                           "doubleclick",
                           "dragclick"]
    def is_button(self):
        return self.type == BUTTON_TYPE

    def is_pressed_only(self):
        return self.pressed and not (self.active or \
                                     self.locked or \
                                     self.scanned)

    def is_text_changing(self):
        if not self.is_modifier() and \
               self.type in [KEYCODE_TYPE,
                             KEYSYM_TYPE,
                             CHAR_TYPE,
                             KEYPRESS_NAME_TYPE,
                             MACRO_TYPE,
                             WORD_TYPE,
                             CORRECTION_TYPE]:
            id = self.id
            if not (id.startswith("F") and id[1:].isdigit()) and \
               not id in set(["LEFT", "RGHT", "UP", "DOWN",
                              "HOME", "END", "PGUP", "PGDN",
                              "INS", "ESC", "MENU",
                              "Prnt", "Pause", "Scroll"]):
                return True
        return False

    def is_return(self):
        id = self.id
        return (id == "RTRN" or
                id == "KPEN")

    def is_separator_cancelling(self):
        """ Should this key cancel pending word separators? """
        return (self.is_correction_key() or
                self.is_return() or
                self.id in set(["SPCE", "TAB",
                                # Don't cancel for Backspace. We want to have
                                # it appear to delete the pending separator.
                                # This way it inserts a space, then immediately
                                # deletes it.
                                # "BKSP",
                                "DELE",
                                "LEFT", "RGHT", "UP", "DOWN",
                                "HOME", "END", "PGUP", "PGDN",
                                "INS", "ESC", "MENU",
                                "Prnt", "Pause", "Scroll"]))

    def get_layer_index(self):
        assert(self.is_layer_button())
        return int(self.id[5:])

    def get_popup_layout(self):
        if self.popup_id:
            return self.find_sublayout(self.popup_id)
        return None

    def can_show_label_popup(self):
        return not self.is_modifier() and \
               not self.is_layer_button() and \
               not self.type is None and \
               bool(self.label_popup)


class RectKeyCommon(KeyCommon):
    """ An abstract class for rectangular keyboard buttons """

    # optional path data for keys with arbitrary shapes
    geometry = None

    # size of rounded corners at 100% round_rect_radius
    chamfer_size = None

    # Optional key_style to override the default theme's style.
    style = None

    # Toggles for what gets drawn.
    show_face = True
    show_border = True
    show_label = True
    show_image = True

    # Allow to display active state, i.e. either latched or locked state.
    # Depending on sticky_behavior the button will still become logically
    # active, it just isn't shown. Used for layer0 buttons, mainly. They don't
    # need to stick out, it's usually obvious when the first layer is active.
    show_active = True

    def __init__(self, id, border_rect):
        KeyCommon.__init__(self)
        self.id = id
        self.colors = {}
        self.context.log_rect = border_rect \
                                if not border_rect is None else Rect()

    def get_id(self):
        return self.id

    def get_svg_id(self):
        return self.svg_id

    def get_state(self):
        state = {}
        state["prelight"]  = self.prelight
        state["pressed"]   = self.pressed
        state["active"]    = self.active
        state["locked"]    = self.locked
        state["scanned"]   = self.scanned
        state["sensitive"] = self.sensitive
        return state

    def draw(self, context = None):
        pass

    def align_label(self, label_size, key_size, ltr = True):
        """ returns x- and yoffset of the aligned label """
        label_x_align = self.label_x_align
        label_y_align = self.label_y_align
        if not ltr:  # right to left script?
            label_x_align = 1.0 - label_x_align
        xoffset = label_x_align * (key_size[0] - label_size[0])
        yoffset = label_y_align * (key_size[1] - label_size[1])
        return xoffset, yoffset

    def align_secondary_label(self, label_size, key_size, ltr = True):
        """ returns x- and yoffset of the aligned label """
        label_x_align = 0.97
        label_y_align = 0.0
        if not ltr:  # right to left script?
            label_x_align = 1.0 - label_x_align
        xoffset = label_x_align * (key_size[0] - label_size[0])
        yoffset = label_y_align * (key_size[1] - label_size[1])
        return xoffset, yoffset

    def align_popup_indicator(self, label_size, key_size, ltr = True):
        """ returns x- and yoffset of the aligned label """
        label_x_align = 1.0
        label_y_align = self.label_y_align
        if not ltr:  # right to left script?
            label_x_align = 1.0 - label_x_align
        xoffset = label_x_align * (key_size[0] - label_size[0])
        yoffset = label_y_align * (key_size[1] - label_size[1])
        return xoffset, yoffset

    def get_style(self):
        if not self.style is None:
            return self.style
        return config.theme_settings.key_style

    def get_stroke_width(self):
        return config.theme_settings.key_stroke_width / 100.0

    def get_stroke_gradient(self):
        return config.theme_settings.key_stroke_gradient / 100.0

    def get_light_direction(self):
        return config.theme_settings.key_gradient_direction * pi / 180.0

    def get_fill_color(self):
        return self._get_color("fill")

    def get_stroke_color(self):
        return self._get_color("stroke")

    def get_label_color(self):
        return self._get_color("label")

    def get_secondary_label_color(self):
        return self._get_color("secondary-label")

    def get_dwell_progress_color(self):
        return self._get_color("dwell-progress")

    def get_dwell_progress_canvas_rect(self):
        rect = self.get_label_rect().inflate(0.5)
        return self.context.log_to_canvas_rect(rect)

    def _get_color(self, element):
        color_key = (element, self.prelight, self.pressed,
                              self.active, self.locked,
                              self.sensitive, self.scanned)
        rgba = self.colors.get(color_key)
        if not rgba:
            if self.color_scheme:
                rgba = self.color_scheme.get_key_rgba(self, element)
            elif element == "label":
                rgba = [0.0, 0.0, 0.0, 1.0]
            else:
                rgba = [1.0, 1.0, 1.0, 1.0]
            self.colors[color_key] = rgba
        return rgba

    def get_fullsize_rect(self):
        """ Get bounding box of the key at 100% size in logical coordinates """
        return LayoutItem.get_rect(self)

    def get_canvas_fullsize_rect(self):
        """ Get bounding box of the key at 100% size in canvas coordinates """
        return self.context.log_to_canvas_rect(self.get_fullsize_rect())

    def get_unpressed_rect(self):
        """
        Get bounding box in logical coordinates.
        Just the relatively static unpressed rect withough fake key action.
        """
        rect = self.get_fullsize_rect()
        return self._apply_key_size(rect)

    def get_rect(self):
        """ Get bounding box in logical coordinates """
        return self.get_sized_rect()

    def get_sized_rect(self, horizontal = None):
        rect = self.get_fullsize_rect()

        # fake physical key action
        if self.pressed:
            dx, dy, dw, dh = self.get_pressed_deltas()
            rect.x += dx
            rect.y += dy
            rect.w += dw
            rect.h += dh

        return self._apply_key_size(rect, horizontal)

    @staticmethod
    def _apply_key_size(rect, horizontal = None):
        """ shrink keys to key_size """
        scale = (1.0 - config.theme_settings.key_size / 100.0) * 0.5
        bx = rect.w * scale
        by = rect.h * scale

        if horizontal is None:
            horizontal = rect.h < rect.w

        if horizontal:
            # keys with aspect > 1.0, e.g. space, shift
            bx = by
        else:
            # keys with aspect < 1.0, e.g. click, move, number block + and enter
            by = bx

        return rect.deflate(bx, by)

    def get_pressed_deltas(self):
        """
        dx, dy, dw, dh for fake physical key action of pressed keys.
        Logical coordinate system.
        """
        key_style = self.get_style()
        if key_style == "gradient":
            k = 0.2
        elif key_style == "dish":
            k = 0.45
        else:
            k = 0.0
        return k, 2*k, 0.0, 0.0

    def get_label_rect(self, rect = None):
        """ Label area in logical coordinates """
        if rect is None:
            rect = self.get_rect()
        style = self.get_style()
        if style == "dish":
            stroke_width  = self.get_stroke_width()
            border_x, border_y = config.DISH_KEY_BORDER
            border_x *= stroke_width
            border_y *= stroke_width
            rect = rect.deflate(border_x, border_y)
            rect.y -= config.DISH_KEY_Y_OFFSET * stroke_width
            return rect
        else:
            return rect.deflate(*self.label_margin)

    def get_canvas_label_rect(self):
        log_rect = self.get_label_rect()
        return self.context.log_to_canvas_rect(log_rect)

    def get_border_path(self):
        """ Original path including border in logical coordinates. """
        return self.geometry.get_full_size_path()

    def get_path(self):
        """
        Path of the key geometry in logical coordinates.
        Key size and fake press movement are applied.
        """
        offset_x, offset_y, size_x, size_y = self.get_key_offset_size()
        return self.geometry.get_transformed_path(offset_x, offset_y,
                                                  size_x, size_y)

    def get_canvas_border_path(self):
        path = self.get_border_path()
        return self.context.log_to_canvas_path(path)

    def get_canvas_path(self):
        path = self.get_path()
        return self.context.log_to_canvas_path(path)

    def get_hit_path(self):
        return self.get_canvas_border_path()

    def get_chamfer_size(self, rect = None):
        """ Max size of the rounded corner areas in logical coordinates. """
        if not self.chamfer_size is None:
            return self.chamfer_size
        if not rect:
            if self.geometry:
                rect = self.get_border_path().get_bounds()
            else:
                rect = self.get_rect()
        return min(rect.w, rect.h) * 0.5

    def get_key_offset_size(self, geometry = None):
        size_x = size_y = config.theme_settings.key_size / 100.0
        offset_x = offset_y = 0.0

        if self.pressed:
            offset_x, offset_y, dw, dh = self.get_pressed_deltas()
            if dw != 0.0 or dh != 0.0:
                if geometry is None:
                    geometry = self.geometry
                dw, dh = geometry.scale_log_to_size((dw, dh))
                size_x += dw * 0.5
                size_y += dh * 0.5

        return offset_x, offset_y, size_x, size_y

    def get_canvas_polygons(self, geometry,
                          offset_x, offset_y, size_x, size_y,
                          radius_pct, chamfer_size):
        path = geometry.get_transformed_path(offset_x, offset_y, size_x, size_y)
        canvas_path = self.context.log_to_canvas_path(path)
        polygons = list(canvas_path.iter_polygons())
        polygon_paths = \
            [polygon_to_rounded_path(p, radius_pct, chamfer_size) \
            for p in polygons]
        return polygons, polygon_paths


class InputlineKeyCommon(RectKeyCommon):
    """ An abstract class for InputLine keyboard buttons """

    line = ""
    word_infos = None
    cursor = 0

    def __init__(self, name, border_rect):
        RectKeyCommon.__init__(self, name, border_rect)

    def get_label(self):
        return ""


class KeyGeometry:
    """
    Full description of a key's shape.

    This class generates path variants for a given key_size by path
    interpolation. This allows for key_size dependent shape changes,
    controlled solely by a SVG layout file. See 'Return' key in
    'Full Keyboard' layout for an example.
    """

    path0 = None          # KeyPath at 100% size
    path1 = None          # KepPath at 50% size, optional

    @staticmethod
    def from_paths(paths):
        assert(len(paths) >= 1)

        path0 = paths[0]
        path1 = None
        if len(paths) >= 2:
            path1 = paths[1]

            # Equal number of path segments?
            if len(path0.segments) != len(path1.segments):
                raise ValueError(
                    "paths to interpolate differ in number of segments "
                    "({} vs. {})" \
                        .format(len(path0.segments), len(path1.segments)))

            # Same operations in all path segments?
            for i in range(len(path0.segments)):
                op0, coords0 = path0.segments[i]
                op1, coords1 = path1.segments[i]
                if op0 != op1:
                    raise ValueError(
                        "paths to interpolate have different operations "
                        "at segment {} (op. {} vs. op. {})" \
                            .format(i, op0, op1))

        geometry = KeyGeometry()
        geometry.path0 = path0
        geometry.path1 = path1
        return geometry

    @staticmethod
    def from_rect(rect):
        geometry = KeyGeometry()
        geometry.path0 = KeyPath.from_rect(rect)
        return geometry

    def get_transformed_path(self, offset_x = 0.0, offset_y = 0.0,
                             size_x = 1.0, size_y = 1.0):
        """
        Everything in the logical coordinate system.
        size: 1.0 => path0, 0.5 => path1
        """
        path0 = self.path0
        path1 = self.path1
        if path1:
            pos_x = (1 - size_x) * 2.0
            pos_y = (1 - size_y) * 2.0
            return path0.linint(path1, pos_x, pos_y, offset_x, offset_y)
        else:
            r0 = self.get_full_size_bounds()
            r1 = self.get_half_size_bounds()
            rect = r1.inflate((size_x - 0.5) * (r0.w - r1.w),
                              (size_y - 0.5) * (r0.h - r1.h))
            rect.x += offset_x
            rect.y += offset_y
            return path0.fit_in_rect(rect)

    def get_full_size_path(self):
        return self.path0

    def get_full_size_bounds(self):
        """
        Bounding box at size 1.0.
        """
        return self.path0.get_bounds()

    def get_half_size_bounds(self):
        """
        Bounding box at size 0.5.
        """
        path1 = self.path1
        if path1:
            rect = path1.get_bounds()
        else:
            rect = self.path0.get_bounds()
            if rect.h < rect.w:
                dx = dy = rect.h * 0.25
            else:
                dy = dx = rect.w * 0.25
            rect = rect.deflate(dx, dy)
        return rect

    def scale_log_to_size(self, v):
        """ Scale from logical distances to key size. """
        r0 = self.get_full_size_bounds()
        r1 = self.get_half_size_bounds()
        log_h = (r0.h - r1.h) * 2.0
        log_w = (r0.w - r1.w) * 2.0
        return (v[0] / log_h,
                v[1] / log_w)

    def scale_size_to_log(self, v):
        """ Scale from logical distances to key size. """
        r0 = self.get_full_size_bounds()
        r1 = self.get_half_size_bounds()
        log_h = (r0.h - r1.h) * 2.0
        log_w = (r0.w - r1.w) * 2.0
        return (v[0] * log_h,
                v[1] * log_w)


class KeyPath:
    """
    Cairo-friendly path description for non-rectangular keys.
    Can handle straight line-loops/polygons, but not arcs and splines.
    """
    (
        MOVE_TO,
        LINE_TO,
        CLOSE_PATH,
    ) = range(3)

    _last_abs_pos = (0.0, 0.0)
    _bounds = None           # cached bounding box

    def __init__(self):
        self.segments = []   # normalized list of path segments (all absolute)

    @staticmethod
    def from_svg_path(path_str):
        path = KeyPath()
        path.append_svg_path(path_str)
        return path

    @staticmethod
    def from_rect(rect):
        x0 = rect.x
        y0 = rect.y
        x1 = rect.right()
        y1 = rect.bottom()
        path = KeyPath()
        path.segments = [[KeyPath.MOVE_TO, [x0, y0]],
                         [KeyPath.LINE_TO, [x1, y0, x1, y1, x0, y1]],
                         [KeyPath.CLOSE_PATH, []]]
        path._bounds = rect.copy()
        return path

    _svg_path_pattern = re.compile("([+-]?[0-9.]+)")

    def copy(self):
        result = KeyPath()
        for op, coords in self.segments:
            result.segments.append([op, coords[:]])
        return result

    def append_svg_path(self, path_str):
        """
        Append a SVG path data string to the path.

        Doctests:
        # absolute move_to command
        >>> p = KeyPath.from_svg_path("M 100 200 120 -220")
        >>> print(p.segments)
        [[0, [100.0, 200.0]], [1, [120.0, -220.0]]]

        # relative move_to command
        >>> p = KeyPath.from_svg_path("m 100 200 10 -10")
        >>> print(p.segments)
        [[0, [100.0, 200.0]], [1, [110.0, 190.0]]]

        # relative move_to and close_path segments
        >>> p = KeyPath.from_svg_path("m 100 200 10 -10 z")
        >>> print(p.segments)
        [[0, [100.0, 200.0]], [1, [110.0, 190.0]], [2, []]]

        # spaces and commas and are optional where possible
        >>> p = KeyPath.from_svg_path("m100,200 10-10z")
        >>> print(p.segments)
        [[0, [100.0, 200.0]], [1, [110.0, 190.0]], [2, []]]
        """

        cmd_str = ""
        coords = []
        tokens = self._tokenize_svg_path(path_str)
        for token in tokens:
            try:
                val = float(token)   # raises value error
                coords.append(val)
            except ValueError:
                if token.isalpha():
                    if cmd_str:
                        self.append_command(cmd_str, coords)
                    cmd_str = token
                    coords = []

                elif token == ",":
                    pass

                else:
                    raise ValueError(
                          "unexpected token '{}' in svg path data" \
                          .format(token))

        if cmd_str:
            self.append_command(cmd_str, coords)

    def append_command(self, cmd_str, coords):
        """
        Append a single command and it's coordinate data to the path.

        Doctests:
        # first lowercase move_to position is absolute
        >>> p = KeyPath()
        >>> p.append_command("m", [100, 200])
        >>> print(p.segments)
        [[0, [100, 200]]]

        # move_to segments become line_to segments after the first position
        >>> p = KeyPath()
        >>> p.append_command("M", [100, 200, 110, 190])
        >>> print(p.segments)
        [[0, [100, 200]], [1, [110, 190]]]

        # further lowercase move_to positions are relative, must become absolute
        >>> p = KeyPath()
        >>> p.append_command("m", [100, 200, 10, -10, 10, -10])
        >>> print(p.segments)
        [[0, [100, 200]], [1, [110, 190, 120, 180]]]

        # further lowercase segments must still be become absolute
        >>> p = KeyPath()
        >>> p.append_command("m", [100, 200, 10, -10, 10, -10])
        >>> p.append_command("l", [1, -1, 1, -1])
        >>> print(p.segments)
        [[0, [100, 200]], [1, [110, 190, 120, 180]], [1, [121, 179, 122, 178]]]
        """

        # Convert lowercase segments from relative to absolute coordinates.
        if cmd_str in ("m", "l"):

            # Don't convert the very first coordinate, it is already absolute.
            if self.segments:
                start = 0
                x, y = self._last_abs_pos
            else:
                start = 2
                x, y = coords[0], coords[1]

            for i in range(start, len(coords), 2):
                x += coords[i]
                y += coords[i+1]
                coords[i]   = x
                coords[i+1] = y

        cmd = cmd_str.lower()
        if cmd == "m":
            self.segments.append([self.MOVE_TO, coords[:2]])
            if len(coords) > 2:
                self.segments.append([self.LINE_TO, coords[2:]])

        elif cmd == "l":
            self.segments.append([self.LINE_TO, coords])

        elif cmd == "z":
            self.segments.append([self.CLOSE_PATH, []])

        # remember last absolute position
        if len(coords) >= 2:
            self._last_abs_pos = coords[-2:]

    @staticmethod
    def _tokenize_svg_path(path_str):
        """
        Split SVG path date into command and coordinate tokens.

        Doctests:
        >>> KeyPath._tokenize_svg_path("m 10,20")
        ['m', '10', ',', '20']
        >>> KeyPath._tokenize_svg_path("   m   10  , \\n  20 ")
        ['m', '10', ',', '20']
        >>> KeyPath._tokenize_svg_path("m 10,20 30,40 z")
        ['m', '10', ',', '20', '30', ',', '40', 'z']
        >>> KeyPath._tokenize_svg_path("m10,20 30,40z")
        ['m', '10', ',', '20', '30', ',', '40', 'z']
        >>> KeyPath._tokenize_svg_path("M100.32 100.09 100. -100.")
        ['M', '100.32', '100.09', '100.', '-100.']
        >>> KeyPath._tokenize_svg_path("m123+23 20,-14L200,200")
        ['m', '123', '+23', '20', ',', '-14', 'L', '200', ',', '200']
        >>> KeyPath._tokenize_svg_path("m123+23 20,-14L200,200")
        ['m', '123', '+23', '20', ',', '-14', 'L', '200', ',', '200']
        """
        tokens = [token.strip() \
                  for token in KeyPath._svg_path_pattern.split(path_str)]
        return [token for token in tokens if token]

    def get_bounds(self):
        bounds = self._bounds
        if bounds is None:
            bounds = self._calc_bounds()
            self._bounds = bounds
        return bounds

    def _calc_bounds(self):
        """
        Compute the bounding box of the path.

        Doctests:
        # Simple move_to path, something inkscape would create.
        >>> p = KeyPath.from_svg_path("m 100,200 10,-10 z")
        >>> print(p.get_bounds())
        Rect(x=100.0 y=190.0 w=10.0 h=10.0)
        """

        try:
            xmin = xmax = self.segments[0][1][0]
            ymin = ymax = self.segments[0][1][1]
        except IndexError:
            return Rect()

        for command in self.segments:
            coords = command[1]
            for i in range(0, len(coords), 2):
                x = coords[i]
                y = coords[i+1]
                if xmin > x:
                    xmin = x
                if xmax < x:
                    xmax = x
                if ymin > y:
                    ymin = y
                if ymax < y:
                    ymax = y

        return Rect(xmin, ymin, xmax - xmin, ymax - ymin)

    def inflate(self, dx, dy = None):
        """
        Returns a new path which is larger by dx and dy on all sides.
        """
        rect = self.get_bounds().inflate(dx, dy)
        return self.fit_in_rect(rect)

    def fit_in_rect(self, rect):
        """
        Scales and translates the path so that rect
        becomes its new bounding box.
        """
        result = self.copy()
        bounds = self.get_bounds()
        scalex = rect.w / bounds.w
        scaley = rect.h / bounds.h
        dorgx, dorgy = bounds.get_center()
        dx = rect.x - (dorgx + (bounds.x - dorgx) * scalex)
        dy = rect.y - (dorgy + (bounds.y - dorgy) * scaley)

        for op, coords in result.segments:
            for i in range(0, len(coords), 2):
                coords[i] = dx + dorgx + (coords[i] - dorgx) * scalex
                coords[i+1] = dy + dorgy + (coords[i+1] - dorgy) * scaley

        return result

    def linint(self, path1, pos_x = 1.0, pos_y = 1.0,
               offset_x = 0.0, offset_y = 0.0):
        """
        Interpolate between self and path1.
        Paths must have the same structure (length and operations).
        pos: 0.0 = self, 1.0 = path1.
        """
        result = self.copy()
        segments = result.segments
        segments1 = path1.segments
        for i in range(len(segments)):
            op, coords = segments[i]
            op1, coords1 = segments1[i]
            for j in range(0, len(coords), 2):
                x = coords[j]
                y = coords[j+1]
                x1 = coords1[j]
                y1 = coords1[j+1]
                dx = x1 - x
                dy = y1 - y
                coords[j] = x + pos_x * dx + offset_x
                coords[j+1] = y + pos_y * dy + offset_y

        return result

    def iter_polygons(self):
        """
        Loop through all independent polygons in the path.
        Can't handle splines and arcs, everything has to
        be polygons from here.
        """
        polygon = []

        for op, coords in self.segments:

            if op == self.LINE_TO:
                polygon.extend(coords)

            elif op == self.MOVE_TO:
                polygon = []
                polygon.extend(coords)

            elif op == self.CLOSE_PATH:
                yield polygon

    def is_point_within(self, point):
        for polygon in self.iter_polygons():
            if self.is_point_in_polygon(polygon, point[0], point[1]):
                return True

    @staticmethod
    def is_point_in_polygon(vertices, x, y):
        c = False
        n = len(vertices)

        try:
            x0 = vertices[n - 2]
            y0 = vertices[n - 1]
        except IndexError:
            return False

        for i in range(0, n, 2):
            x1 = vertices[i]
            y1 = vertices[i+1]
            if (y1 <= y and y < y0 or y0 <= y and y < y1) and \
               (x < (x0 - x1) * (y - y1) / (y0 - y1) + x1):
                c = not c
            x0 = x1
            y0 = y1

        return c





#!/usr/bin/python3

# Copyright © 2015 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import tempfile
import re
import subprocess
import threading
import unittest
from contextlib import contextmanager, ExitStack

import dbus
from dbus.mainloop.glib import DBusGMainLoop

from Onboard.Version import require_gi_versions
require_gi_versions()
from gi.repository import GLib, Gtk

from Onboard.utils import Rect
from Onboard.LanguageSupport import LanguageDB
from Onboard.definitions import Handle


DBUS_NAME = "org.onboard.Onboard"
DBUS_PATH  = "/org/onboard/Onboard/Keyboard"
DBUS_IFACE = "org.onboard.Onboard.Keyboard"


class _TestGUIBase(unittest.TestCase):

    class OnboardRemoteInstance:
        def __init__(self, p, bus):
            self.process = p
            self._bus = bus
            self._keyboard = None
            self._keys = []

        def get_keyboard(self):
            if self._keyboard is None:
                proxy = self._bus.get_object(DBUS_NAME, DBUS_PATH)
                self._keyboard = dbus.Interface(proxy, DBUS_IFACE)
            return self._keyboard

        def get_keys(self):
            # if not self._keys:
            if True:  # don't lose sync when switching layers
                class Key:
                    pass
                keyboard = self.get_keyboard()
                for id, extents, labels, states in keyboard.GetKeyState():
                    key = Key()
                    key.id = id
                    key.labels = labels
                    key.screen_rect = Rect(*extents)
                    for name, value in states.items():
                        setattr(key, name, value)
                    self._keys.append(key)

            return self._keys

        def char_to_key_id(self, char):
            if char == " ":
                return "SPCE"
            for key in self.get_keys():
                if char in key.labels.values():
                    return key.id
            else:
                raise ValueError("unknown character '{}'".format(char))
            return ""

        def get_key(self, key_id):
            """ Center of key in screen coordinates """
            for key in self.get_keys():
                if key.id == key_id:
                    return key
            else:
                raise ValueError("unknown key_id '{}'".format(key_id))
            return None

        def get_key_center(self, key_id):
            """ Center of key in screen coordinates """
            key = self.get_key(key_id)
            x, y = key.screen_rect.get_center()
            return int(x), int(y)

    class TextWindow:
        """ Little window with a text entry to capture Onboard's typing in."""

        def __init__(self):
            self.loop = GLib.MainLoop()
            self.win = Gtk.Window(default_height=50, default_width=300)
            self.win.connect("delete-event", lambda *x: self.loop.quit())
            self.entry = Gtk.Entry()
            self.win.add(self.entry)
            self.win.show_all()
            self._running = True
            self.process_events()
            #threading.Thread(target=self.run_loop).start()

        def close(self):
            text = self.entry.get_text()
            self.process_events()
            self.loop.quit()
            self.process_events()
            self._running = False
            self.win.destroy()
            self.process_events()
            return text

        def process_events(self):
            while self.loop.get_context().iteration(False):
                pass

        def run_loop(self):
            while self._running:
                self.loop.get_context().iteration()

    def setUp(self):
        self._tmp_dir = tempfile.TemporaryDirectory(prefix="test_onboard_")
        self._dir = self._tmp_dir.name
        self._user_dir = os.path.join(self._dir, "onboard")
        self._model_dir = os.path.join(self._user_dir, "models")

        # setup a fresh dconf database
        self._dconf_db_name = "_onboard_test_db"
        self._dconf_db_fn = os.path.join(self.get_config_home(),
                                         "dconf", self._dconf_db_name)
        self._dconf_profile_fn = os.path.join(self._dir, "dconf_profile")
        self._write_file(self._dconf_profile_fn,
                         ["user-db:" + self._dconf_db_name])
        self._dconf_reset()  # clear everything

        self._bus = dbus.SessionBus(mainloop=DBusGMainLoop())

    def tearDown(self):
        os.remove(self._dconf_db_fn)
        self._tmp_dir.cleanup()

    def assertRectInRange(self, r_result, r_expect, tolerance=0):
        for attr in Rect.attributes:
            with self.subTest(attribute=attr):
                expected = getattr(r_expect, attr)
                result = getattr(r_result, attr)
                self.assertInRange(result,
                                   expected - tolerance,
                                   expected + tolerance,
                                   "Rect." + attr)

    def assertInRange(self, value, begin, end, variable):
        self.assertTrue(begin <= value <= end,
                    "{}={} not in range {}..{}"
                    .format(variable, value, begin, end))

    def _key_center(self, rwin, key_id):
        frame = 5
        rkb = rwin.copy()

        # word suggestion row visible?
        if self._gsettings_get(
                "org.onboard.typing-assistance.word-suggestions", "enabled"):
            hws = (rwin.h - 2 * frame) / 6
            rkb.y += hws
            rkb.h -= hws

        h_key = (rkb.h - 2 * frame) / 5
        w_key = h_key

        if key_id == "move":
            x = rkb.right() - w_key / 2
            y = rkb.y + frame + 1.5 * h_key

        if key_id == "layer1":
            x = rkb.right() - w_key / 2
            y = rkb.y + frame + 3.5 * h_key

        if key_id == "NMLK":
            x = rkb.x + rkb.w * 0.7
            y = rkb.y + frame + 0.5 * h_key

        return int(x), int(y)

    def _get_window_rect(self):
        return self._get_window_landscape_rect()

    def _set_window_rect(self, r):
        self._set_window_landscape_rect(r)

    def _get_window_landscape_rect(self):
        return Rect(
            self._gsettings_get("org.onboard.window.landscape", "x"),
            self._gsettings_get("org.onboard.window.landscape", "y"),
            self._gsettings_get("org.onboard.window.landscape", "width"),
            self._gsettings_get("org.onboard.window.landscape", "height"))

    def _set_window_landscape_rect(self, r):
        self._gsettings_set("org.onboard.window.landscape", "x", r.x)
        self._gsettings_set("org.onboard.window.landscape", "y", r.y)
        self._gsettings_set("org.onboard.window.landscape", "width", r.w)
        self._gsettings_set("org.onboard.window.landscape", "height", r.h)

    def _get_window_portrait_rect(self):
        return Rect(
            self._gsettings_get("org.onboard.window.portrait", "x"),
            self._gsettings_get("org.onboard.window.portrait", "y"),
            self._gsettings_get("org.onboard.window.portrait", "width"),
            self._gsettings_get("org.onboard.window.portrait", "height"))

    def _set_window_portrait_rect(self, r):
        self._gsettings_set("org.onboard.window.portrait", "x", r.x)
        self._gsettings_set("org.onboard.window.portrait", "y", r.y)
        self._gsettings_set("org.onboard.window.portrait", "width", r.w)
        self._gsettings_set("org.onboard.window.portrait", "height", r.h)

    def _get_icon_palette_rect(self):
        return self._get_icon_palette_landscape_rect()

    def _set_icon_palette_rect(self, r):
        self._set_window_landscape_rect(r)

    def _get_icon_palette_landscape_rect(self):
        schema = "org.onboard.icon-palette.landscape"
        return Rect(
            self._gsettings_get(schema, "x"),
            self._gsettings_get(schema, "y"),
            self._gsettings_get(schema, "width"),
            self._gsettings_get(schema, "height"))

    def _set_icon_palette_landscape_rect(self, r):
        schema = "org.onboard.icon-palette.landscape"
        self._gsettings_set(schema, "x", r.x)
        self._gsettings_set(schema, "y", r.y)
        self._gsettings_set(schema, "width", r.w)
        self._gsettings_set(schema, "height", r.h)

    def _get_icon_palette_portrait_rect(self):
        schema = "org.onboard.icon-palette.portrait"
        return Rect(
            self._gsettings_get(schema, "x"),
            self._gsettings_get(schema, "y"),
            self._gsettings_get(schema, "width"),
            self._gsettings_get(schema, "height"))

    def _set_icon_palette_portrait_rect(self, r):
        schema = "org.onboard.icon-palette.portrait"
        self._gsettings_set(schema, "x", r.x)
        self._gsettings_set(schema, "y", r.y)
        self._gsettings_set(schema, "width", r.w)
        self._gsettings_set(schema, "height", r.h)

    def _enable_major_features(self, auto_show=True):
        # turn major features on
        self._gsettings_set("org.onboard.auto-show", "enabled", auto_show)
        self._gsettings_set("org.onboard.typing-assistance.word-suggestions",
                            "enabled", True)
        self._gsettings_set("org.onboard.icon-palette", "in-use", True)

        # no dialog on startup
        self._gsettings_set("org.gnome.desktop.interface",
                            "toolkit-accessibility", True)
    def _use_system_defaults(self, use):
        self._gsettings_set("org.onboard", "use-system-defaults", use)


    def _mouse_sweep(self, x0, y0, x1, y1, step=5):
        xd = x1 - x0
        yd = y1 - y0
        n = int(max(abs(xd), abs(yd)) / step)
        for i in range(n):
            x = x0 + xd * i / (n - 1)
            y = y0 + yd * i / (n - 1)
            self._xte("mousemove", int(x), int(y))
            # time.sleep(0.1)

        # The last motion event seems to be only sent for the core
        # pointer device. -> Trigger a redundant motion event with the last
        # position to make sure there is an event coming from the XTest
        # slave device too. Else handles won't move all the way in the
        # move/resize tests.
        self._xte("mousemove", int(x1), int(y1))

    def _click_string(self, remote_instance, s):
        for c in s:
            key_id = remote_instance.char_to_key_id(c)
            x, y = remote_instance.get_key_center(key_id)
            self._xte("mousemove", x, y)
            self._xte("mouseclick", 1)
            time.sleep(0.3)

    def _click_key(self, remote_instance, key_id):
        x, y = remote_instance.get_key_center(key_id)
        self._xte("mousemove", x, y)
        self._xte("mouseclick", 1)
        time.sleep(0.3)

    @contextmanager
    def _run_onboard(self, params=[], env={},
                     capture_output=False, print_output=False):
        with self._run_onboard_script("./onboard", params, env,
                                      capture_output, print_output) as p:
            self._wait_name_owner_changed(self._bus, DBUS_NAME)
            time.sleep(1)
            instance = self.OnboardRemoteInstance(p, self._bus)
            yield instance
            p.terminate()

    @contextmanager
    def _run_onboard_settings(self, params=[], env={},
                              capture_output=False, print_output=False,
                              delay=2.0):
        with self._run_onboard_script("./onboard-settings",
                                      params, env,
                                      capture_output, print_output) as p:
            time.sleep(delay)
            yield p
            p.terminate()
            # self._xte("key", "Escape")

    @contextmanager
    def _run_onboard_script(self, command, params=[], _env={},
                            capture_output=False, print_output=False):
        try:
            env = dict(os.environ)
            env["DCONF_PROFILE"] = self._dconf_profile_fn
            env["XDG_DATA_HOME"] = self._dir
            env["LANG"] = "en_US.UTF-8"
            env.update(_env)

            cmd_line = [command]
            if params:
                cmd_line += params

            if capture_output:
                stdout_opt = subprocess.PIPE
                stderr_opt = subprocess.PIPE
            elif print_output:
                stdout_opt = subprocess.PIPE
                stderr_opt = subprocess.STDOUT
            else:
                stdout_opt = None
                stderr_opt = None

            p = subprocess.Popen(cmd_line, env=env,
                                 stdout=stdout_opt,
                                 stderr=stderr_opt,
                                 close_fds=True,
                                 universal_newlines=True,
                                 )
            yield p

        finally:
            if capture_output:
                p.stdout_lines = [line.replace("\n", "") for line in p.stdout]
                p.stderr_lines = [line.replace("\n", "") for line in p.stderr]

                # prevent ResourceWarning:
                # unclosed file <_io.BufferedReader name=12> testMethod()
                p.stderr.close()
                p.stdout.close()

            elif print_output:
                # filter out annoying Gtk deprecation warnings
                # we cannot remove yet
                blacklist = [
                    "is deprecated and shouldn't be used anymore",
                    "builder.add_from_file",
                    "GtkDialog mapped without a transient parent",
                ]
                for line in p.stdout:
                    line = line.replace("\n", "")
                    if line and \
                       not any(s in line for s in blacklist):
                        print("stderr:", line, file=sys.stderr)

                # prevent ResourceWarning:
                # unclosed file <_io.BufferedReader name=12> testMethod()
                p.stdout.close()

    def _system_gsettings_set(self, schema, key, value):
        self._gsettings_set_with_env(schema, key, value, None)

    def _gsettings_set(self, schema, key, value):
        env = dict(os.environ)
        env["DCONF_PROFILE"] = self._dconf_profile_fn
        env["LANG"] = "en_US.UTF-8"
        self._gsettings_set_with_env(schema, key, value, env)

    def _gsettings_set_with_env(self, schema, key, value, env):
        if value is True:
            valstr = "true"
        elif value is False:
            valstr = "false"
        else:
            valstr = str(value)

        p = subprocess.Popen(["gsettings", "set",
                              schema, key, valstr], env=env)
        p.wait()

    def _system_gsettings_get(self, schema, key):
        return self._gsettings_get_with_env(schema, key, None)

    def _gsettings_get(self, schema, key):
        env = dict(os.environ)
        env["DCONF_PROFILE"] = self._dconf_profile_fn
        env["LANG"] = "en_US.UTF-8"
        return self._gsettings_get_with_env(schema, key, env)

    def _gsettings_get_with_env(self, schema, key, env):
        valstr = subprocess.check_output(["gsettings", "get",
                                         schema, key], env=env)
        valstr = valstr.decode("UTF-8").replace("\n", "")
        if valstr == "true":
            value = True
        elif valstr == "false":
            value = False
        else:
            try:
                value = int(valstr)
            except ValueError:
                try:
                    value = float(valstr)
                except ValueError:
                    value = valstr
        return value

    def _dconf_reset(self):
        env = dict(os.environ)
        env["DCONF_PROFILE"] = self._dconf_profile_fn
        env["LANG"] = "en_US.UTF-8"
        subprocess.check_call(["dconf", "reset", "-f", "/"], env=env)

    def _xte(self, *params):
        env = dict(os.environ)
        env["LANG"] = "en_US.UTF-8"
        command = ["xte", " ".join(str(p) for p in params)]
        # print(command, file=sys.stderr)
        subprocess.check_call(command, env=env)

    def _get_numlock_state(self):
        output = subprocess.check_output(["numlockx", "status"]).decode()
        if " off" in output:
            return False
        elif " on" in output:
            return True
        else:
            self.assertTrue(False)
        return None

    def _set_numlock_state(self, on):
        subprocess.check_call(["numlockx", "on" if on else "off"])

    def _rotate_screen(self, rotation):
        output = subprocess.check_output(
            ['/bin/bash', '-c', 'xrandr | grep primary | cut -d" " -f1'])
        output = output.decode().replace("\n", "")
        if rotation == "landscape":
            subprocess.check_call(
                ['xrandr', "--output", output, "--rotate", "normal"])
        elif rotation == "portrait":
            subprocess.check_call(
                ['xrandr', "--output", output, "--rotate", "left"])

    @staticmethod
    def _wait_name_owner_changed(bus, name):

        def on_name_owner_changed(name, old, new):
            if not old:
                loop.quit()

        bus.add_signal_receiver(on_name_owner_changed,
                                "NameOwnerChanged",
                                dbus.BUS_DAEMON_IFACE,
                                arg0=name)
        loop = GLib.MainLoop()
        loop.run()

    @staticmethod
    def get_config_home(file=None):
        """
        User specific config directory.
        """
        path = os.environ.get("XDG_CONFIG_HOME")
        if path and not os.path.isabs(path):
            path = None

        if not path:
            path = os.path.join(os.path.expanduser("~"), ".config")

        if file:
            path = os.path.join(path, file)

        return path

    @staticmethod
    def _write_file(fn, lines):
        with open(fn, mode="w", encoding="UTF-8") as f:
            for l in lines:
                f.write(l)

    @staticmethod
    def _touch(fn, size):
        with open(fn, mode="w") as f:
            if size:
                f.write("*" * size)


class TestWindowHandling(_TestGUIBase):

    def test_keyboard_moving_remembered_after_restart(self):
        r = self._get_window_rect()
        dx = 100
        dy = 200
        with self._run_onboard():
            x, y = self._key_center(r, "move")
            self._xte("mousemove", x, y)
            self._xte("mousedown", 1)
            time.sleep(0.5)
            self._mouse_sweep(x, y, x + dx, y + dy)
            time.sleep(0.5)
            self._xte("mouseup", 1)
            time.sleep(0.5)

        k = 20  # ignore a few lost pixels due to drag threshold and xte moving
        r_expect = r.offset(dx, dy)
        r_result = self._get_window_rect()
        self.assertEqual(r_expect.get_size(), r_result.get_size())
        self.assertTrue(r_expect.x - k <= r_result.x <= r_expect.x,
                        "expected={} result={}".format(r_expect.x, r_result.x))
        self.assertTrue(r_expect.y - k <= r_result.y <= r_expect.y,
                        "expected={} result={}".format(r_expect.y, r_result.y))

        # rect must not change after next restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_window_rect()))

        # rect must not change after another restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_window_rect()))

    def test_keyboard_resizing_remembered_after_restart(self):
        r = self._get_window_rect()
        dx = 100
        dy = 200
        dx_handle = 5
        dy_handle = 5
        with self._run_onboard():
            x, y = r.right() - dx_handle, r.bottom() - dy_handle
            self._xte("mousemove", x, y)
            self._xte("mousedown", 1)
            time.sleep(0.5)
            self._mouse_sweep(x, y, x + dx, y + dy)
            time.sleep(0.5)
            self._xte("mouseup", 1)
            time.sleep(0.5)

        k = 20  # ignore a few lost pixels due to drag threshold and xte moving
        r_expect = r.copy()
        r_expect.w = r.w + dx
        r_expect.h = r.h + dy
        r_result = self._get_window_rect()

        self.assertEqual(r_expect.left_top(), r_result.left_top())
        self.assertTrue(r_expect.w - k <= r_result.w <= r_expect.w,
                        "expected={} result={}".format(r_expect.w, r_result.w))
        self.assertTrue(r_expect.h - k <= r_result.h <= r_expect.h,
                        "expected={} result={}".format(r_expect.h, r_result.h))

        # rect must not change after next restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_window_rect()))

        # rect must not change after another restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_window_rect()))

    def test_icon_palette_moving_remembered_after_restart(self):
        self._gsettings_set("org.onboard", "start-minimized", True)
        self._gsettings_set("org.onboard.icon-palette", "in-use", True)

        r = self._get_icon_palette_rect()
        dx = 100
        dy = 200
        with self._run_onboard():
            x, y = (int(val) for val in r.get_center())
            self._xte("mousemove", x, y)
            self._xte("mousedown", 1)
            time.sleep(0.5)
            self._mouse_sweep(x, y, x + dx, y + dy)
            time.sleep(0.5)
            self._xte("mouseup", 1)
            time.sleep(0.5)

        k = 10  # ignore a few lost pixels due to drag threshold and xte moving
        r_expect = r.offset(dx, dy)
        r_result = self._get_icon_palette_rect()
        self.assertEqual(r_expect.get_size(), r_result.get_size())
        self.assertTrue(r_expect.x - k <= r_result.x <= r_expect.x,
                        "expected={} result={}".format(r_expect.x, r_result.x))
        self.assertTrue(r_expect.y - k <= r_result.y <= r_expect.y,
                        "expected={} result={}".format(r_expect.y, r_result.y))

        # rect must not change after next restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_icon_palette_rect()))

        # rect must not change after another restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_icon_palette_rect()))

    def test_icon_palette_resizing_remembered_after_restart(self):
        self._gsettings_set("org.onboard", "start-minimized", True)
        self._gsettings_set("org.onboard.icon-palette", "in-use", True)

        r = self._get_icon_palette_rect()
        dx = 100
        dy = 200
        dx_handle = 5
        dy_handle = 5
        with self._run_onboard():
            x, y = r.right() - dx_handle, r.bottom() - dy_handle
            self._xte("mousemove", x, y)
            self._xte("mousedown", 1)
            time.sleep(0.5)
            self._mouse_sweep(x, y, x + dx, y + dy)
            time.sleep(0.5)
            self._xte("mouseup", 1)
            time.sleep(0.5)

        k = 10  # ignore a few lost pixels due to drag threshold and xte moving
        r_expect = r.copy()
        r_expect.w = r.w + dx
        r_expect.h = r.h + dy
        r_result = self._get_icon_palette_rect()

        self.assertEqual(r_expect.left_top(), r_result.left_top())
        self.assertTrue(r_expect.w - k <= r_result.w <= r_expect.w,
                        "expected={} result={}".format(r_expect.w, r_result.w))
        self.assertTrue(r_expect.h - k <= r_result.h <= r_expect.h,
                        "expected={} result={}".format(r_expect.h, r_result.h))

        # rect must not change after next restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_icon_palette_rect()))

        # rect must not change after another restart
        with self._run_onboard():
            pass
        self.assertEqual(str(r_result), str(self._get_icon_palette_rect()))

    @unittest.skip("unfinished")
    def test_screen_rotation(self):
        self._enable_major_features()
        self._gsettings_set("org.onboard.icon-palette", "in-use", True)
        self._gsettings_set("org.onboard.icon-palette.landscape", "x", 600)
        self._gsettings_set("org.onboard.icon-palette.landscape", "y", 200)
        self._gsettings_set("org.onboard.window.landscape", "x", 200)
        self._gsettings_set("org.onboard.window.landscape", "y", 300)

        self._gsettings_set("org.onboard.window.portrait", "x", 200)
        self._gsettings_set("org.onboard.window.portrait", "y", 600)

        self._rotate_screen("landscape")

        dx = 100
        dy = 100
        tolerance = 15
        r_before_landscape = self._get_window_landscape_rect()
        r_before_portrait = self._get_window_portrait_rect()
        with self._run_onboard():
            self._move_handle_rel(Handle.NORTH_WEST, r_before_landscape,
                                  dx, dy)
            self._move_handle_rel(Handle.SOUTH_EAST, r_before_landscape,
                                  dx, dy)
            time.sleep(1.0)

            self._rotate_screen("portrait")
            time.sleep(5)

            self._move_handle_rel(Handle.NORTH_WEST, r_before_portrait, dx, dy)
            self._move_handle_rel(Handle.SOUTH_EAST, r_before_portrait, dx, dy)
            time.sleep(1.0)

            self._rotate_screen("landscape")
            time.sleep(5)

        time.sleep(0.5)

        # print (r, file=sys.stderr)
        # print (r.inflate(dx, dy), file=sys.stderr)
        # print (self._get_window_landscape_rect(), file=sys.stderr)
        self.assertRectInRange(self._get_window_landscape_rect(),
                               r_before_landscape.inflate(dx, dy), tolerance)
        self.assertRectInRange(self._get_window_portrait_rect(),
                               r_before_portrait.inflate(dx, dy), tolerance)

    def test_icon_palette_resizing(self):
        self._gsettings_set("org.onboard", "start-minimized", True)
        self._gsettings_set("org.onboard.icon-palette", "in-use", True)
        self._gsettings_set("org.onboard.icon-palette.landscape", "x", 300)
        self._gsettings_set("org.onboard.icon-palette.landscape", "y", 300)
        self._test_window_resizing(self._get_icon_palette_rect,
                                   self._set_icon_palette_rect)

    def test_keyboard_window_resizing(self):
        self._gsettings_set("org.onboard.window.landscape", "x", 300)
        self._gsettings_set("org.onboard.window.landscape", "y", 300)
        self._test_window_resizing(self._get_window_rect,
                                   self._set_window_rect)

    def _test_window_resizing(self, get_window_rect, set_window_rect):
        dx = 100
        dy = 100
        tolerance = 15  # there's some loss due to drag-threshold and others
        r_default = get_window_rect()

        with self.subTest(comment="corners1"):
            set_window_rect(r_default)

            r = get_window_rect()
            with self._run_onboard():
                self._move_handle_rel(Handle.NORTH_WEST, r, dx, dy)
                self._move_handle_rel(Handle.SOUTH_EAST, r, dx, dy)

            self.assertRectInRange(get_window_rect(),
                                r.inflate(dx, dy), tolerance)

            r = get_window_rect()
            with self._run_onboard():
                self._move_handle_rel(Handle.NORTH_WEST, r, -dx, -dy)
                self._move_handle_rel(Handle.SOUTH_EAST, r, -dx, -dy)

            self.assertRectInRange(get_window_rect(),
                                r.deflate(dx, dy), tolerance)

        with self.subTest(comment="corners2"):
            set_window_rect(r_default)

            r = get_window_rect()
            with self._run_onboard():
                self._move_handle_rel(Handle.SOUTH_WEST, r, dx, dy)
                self._move_handle_rel(Handle.NORTH_EAST, r, dx, dy)

            self.assertRectInRange(get_window_rect(),
                                r.inflate(dx, dy), tolerance)

            r = get_window_rect()
            with self._run_onboard():
                self._move_handle_rel(Handle.SOUTH_WEST, r, -dx, -dy)
                self._move_handle_rel(Handle.NORTH_EAST, r, -dx, -dy)

            self.assertRectInRange(get_window_rect(),
                                r.deflate(dx, dy), tolerance)

        with self.subTest(comment="edges"):
            set_window_rect(r_default)
            r = get_window_rect()
            with self._run_onboard():
                self._move_handle_rel(Handle.NORTH, r, dx, dy)
                self._move_handle_rel(Handle.EAST, r, dx, dy)
                self._move_handle_rel(Handle.SOUTH, r, dx, dy)
                self._move_handle_rel(Handle.WEST, r, dx, dy)

            self.assertRectInRange(get_window_rect(),
                                r.inflate(dx, dy), tolerance)

            r = get_window_rect()
            with self._run_onboard():
                self._move_handle_rel(Handle.NORTH, r, -dx, -dy)
                self._move_handle_rel(Handle.EAST, r, -dx, -dy)
                self._move_handle_rel(Handle.SOUTH, r, -dx, -dy)
                self._move_handle_rel(Handle.WEST, r, -dx, -dy)

            self.assertRectInRange(get_window_rect(),
                                r.deflate(dx, dy), tolerance)

    def _move_handle_rel(self, handle, r, dx, dy):
        x0, y0, x1, y1 = self._get_handle_move_vector(handle, r, dx, dy)
        self._xte("mousemove", x0, y0)
        self._xte("mousedown", 1)
        time.sleep(0.3)
        self._mouse_sweep(x0, y0, x1, y1)
        self._xte("mouseup", 1)
        time.sleep(0.3)

    def _get_handle_move_vector(self, handle, r, dx, dy):
        handle_offset = 4
        if handle in [Handle.NORTH,
                      Handle.NORTH_WEST,
                      Handle.NORTH_EAST]:
            y0 = r.y + handle_offset
            y1 = y0 - dy
        if handle in [Handle.WEST,
                      Handle.NORTH_WEST,
                      Handle.SOUTH_WEST]:
            x0 = r.x + handle_offset
            x1 = x0 - dx
        if handle in [Handle.EAST,
                      Handle.NORTH_EAST,
                      Handle.SOUTH_EAST]:
            x0 = r.right() - handle_offset
            x1 = x0 + dx
        if handle in [Handle.SOUTH,
                      Handle.SOUTH_WEST,
                      Handle.SOUTH_EAST]:
            y0 = r.bottom() - handle_offset
            y1 = y0 + dy

        if handle in [Handle.NORTH,
                      Handle.SOUTH]:
            x0 = int(r.get_center()[0])
            x1 = x0
        if handle in [Handle.WEST,
                      Handle.EAST]:
            y0 = int(r.get_center()[1])
            y1 = y0

        return x0, y0, x1, y1


class TestKeys(_TestGUIBase):

    @unittest.skip("freezes too easily")
    def test_modifier_unlatching(self):
        tests = (
            [
                "Small",
                "[LFSH]abc ",
                "Abc "
            ],
            [
                "Compact",
                "[LFSH]abc [RTSH]def",
                "Abc Def"
            ],
            [
                "Full Keyboard",
                "[LFSH]abc [RTSH]def",
                "Abc Def"
            ],
            [
                "Phone",
                "[LFSH]abc ",
                "Abc "
            ],
            [
                "Grid",
                "[LFSH]abc ",
                "Abc "
            ],
        )

        self._enable_major_features(auto_show=False)
        self._use_system_defaults(False)

        for layout, challenge, expectation in tests:
            self._gsettings_set("org.onboard", "layout", layout)
            with self.subTest(layout=layout, challenge=challenge):
                tw = self.TextWindow()
                with self._run_onboard() as instance:
                    groups = [m.groups() for m
                              in re.finditer("(?: \[ ([^\]]*) ) |"
                                             "(?: \] ([^\[]*) )",
                                             challenge, re.VERBOSE)]
                    for key_id, text in groups:
                        if key_id:
                            self._click_key(instance, key_id)
                        else:
                            self._click_string(instance, text)
                        tw.process_events()

                text = tw.close()
                time.sleep(0.5)
                self.assertEqual(text, expectation)

    def test_numlock_state_on_exit(self):
        self._enable_major_features()

        tests = [
            [
                "NMLK off state must be preserved after exit",  # comment
                0.0,     # sticky_key_release_delay 0.0=off, !=0 = kiosk mode
                False,   # NMLK state before start
                False,   # toggle NMLK key at runtime
                False,   # NMLK state after exit
            ],
            [
                "NMLK on state must be preserved after exit",
                0.0, True, False, True,
            ],
            [
                "NMLK toggled on must still be on after exit",
                0.0, False, True, True,
            ],
            [
                "NMLK toggled off must still be off after exit",
                0.0, True, True, False,
            ],
            # Currently, if sticky_key_release_delay is != 0.0 we assume we're
            # running in a kiosk setting and reset NMLK after the delay as
            # well as on exit.
            [
                "NMLK toggled on in kiosk mode reverts to off after exit",
                5.0, False, True, False,
            ],
        ]

        old_state = self._get_numlock_state()

        for comment, skrd, active_before, toggle, active_after in tests:
            self._set_numlock_state(active_before)
            self._gsettings_set("org.onboard.keyboard",
                                "sticky-key-release-delay", skrd)

            with self._run_onboard() as instance:
                if toggle:
                    self._click_key(instance, "layer1")
                    self._click_key(instance, "NMLK")
                else:
                    time.sleep(1)

            if not toggle:  # run twice only the first two times
                with self._run_onboard():
                    time.sleep(1)

            self.assertEqual(self._get_numlock_state(), active_after, comment)

        self._set_numlock_state(old_state)


class TestMisc(_TestGUIBase):

    def test_valid_test_environment(self):
        # write to dconf database to make sure it exists
        self._gsettings_set("org.onboard", "use-system-defaults", True)
        self.assertEqual(
            self._gsettings_get("org.onboard", "use-system-defaults"),
            True)
        self.assertTrue(os.path.exists(self._dconf_db_fn))

        # database must be reset to defaults
        self.assertEqual(str(self._get_window_rect()),
                         str(Rect(100, 50, 700, 205)))

    def test_startup_with_c_locale(self):
        self._enable_major_features()

        with self._run_onboard(env={"LANG": "C"}):
            pass

        with self._run_onboard_settings(env={"LANG": "C"},
                                        print_output=True):
            pass

    @staticmethod
    def _get_languages():
        return LanguageDB.get_main_languages()

    def test_startup_with_various_languages_onboard(self):
        self._enable_major_features()

        languages = self._get_languages()
        for language in languages:
            with self.subTest(language=language):
                with self._run_onboard(env={"LANGUAGE": language}):
                    pass

    def test_startup_with_various_languages_onboard_settings(self):
        self._enable_major_features()

        languages = self._get_languages()

        # run onboard-settings for all languages
        k = 4  # max parallel languages
        for i in range(0, len(languages), k):
            langs = languages[i:i + k]
            with self.subTest(languages=langs):
                with ExitStack() as stack:
                    for language in langs:
                        stack.enter_context(
                            self._run_onboard_settings(
                                env={"LANGUAGE": language},
                                print_output=True,
                                delay=0.5))
                    time.sleep(2)

    #@unittest.skip("doesn't clean up very well")
    def test_gnome_high_contrast_themes(self):
        self._enable_major_features()

        # make sure system theme tracking is enabled, should be default
        self.assertTrue(self._gsettings_get("org.onboard",
                                            "system-theme-tracking-enabled"))

        default_gtk_theme = "Ambiance"  # default in Vivid is Adwaita?
        default_onboard_theme = "Nightshade"  # default in Vivid is Adwaita?
        contrast_themes = ["HighContrast",
                           "HighContrastInverse",
                           "LowContrast"]

        # save current gsettings state
        old_theme = self._system_gsettings_get("org.gnome.desktop.interface",
                                               "gtk-theme")

        # set normal theme for startup
        self._system_gsettings_set("org.gnome.desktop.interface",
                                   "gtk-theme", default_gtk_theme)

        # run onboard and switch through gtk-themes
        with self._run_onboard(params=["-d", "info"],
                               capture_output=True) as instance:
            time.sleep(2)
            for contrast_theme in contrast_themes:
                self._system_gsettings_set("org.gnome.desktop.interface",
                                        "gtk-theme", contrast_theme)
                time.sleep(3)

            self._system_gsettings_set("org.gnome.desktop.interface",
                                       "gtk-theme", default_gtk_theme)
            time.sleep(3)

        # process debug output
        lines = [line for line in instance.process.stderr_lines
                 if "Loading theme" in line]
        self.assertEqual(len(lines), len(contrast_themes) + 2)
        self.assertIn(default_onboard_theme, lines[0])
        for i, theme in enumerate(contrast_themes):
            self.assertIn(theme + ".theme", lines[i + 1])
        self.assertIn(default_onboard_theme, lines[-1])

        # undo our gsettings changes
        self._system_gsettings_set("org.gnome.desktop.interface",
                                   "gtk-theme", old_theme)

    def test_running_in_live_cd_environment(self):
        # make sure icon palatte is off (default)
        self.assertFalse(
            self._gsettings_get("org.onboard.icon-palette", "in-use"))

        # make sure status icon is on (default)
        self.assertTrue(
            self._gsettings_get("org.onboard", "show-status-icon"))

        with self._run_onboard(params=["-d", "debug"],
                               env={"RUNNING_UNDER_GDM": "1"},
                               capture_output=True) as instance:
            pass

        # preferences key must have become inaccessible
        lines = [line for line in instance.process.stderr_lines
                 if "RectKey('settings').visible" in line]
        self.assertTrue(
            len(lines) >= 1 and all("False" in line for line in lines))
        self.assertTrue(
            len(lines) >= 1 and not any("True" in line for line in lines))

        # icon palette must be on
        self.assertTrue(
            self._gsettings_get("org.onboard.icon-palette", "in-use"))

        # status icon must be off
        self.assertFalse(
            self._gsettings_get("org.onboard", "show-status-icon"))


class TestNoSetup(unittest.TestCase):

    def test_apt_cache_unmet_onboard(self):
        result = subprocess.check_output(["apt-cache", "unmet", "onboard"])
        result = result.decode("UTF-8")
        self.assertEqual(result, "")

